/****************************************************************************
**
** Copyright (C) 1991-2006 Heinz van Saanen
**
** This file is part of the the function viewer Zhu3D.
**
** This file may be used under the terms of the GNU General Public
** License version 2.0 as published by the Free Software Foundation
** and appearing in the file LICENSE.GPL included in the packaging of
** this file.
**
** This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
** WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
**
****************************************************************************/

#include "aniedit.h"
#include "timers.h"


// Constructor
aniWidget::aniWidget(QWidget *parent) : QWidget(parent) {

	// Set up entire editor
	Ui_aniUI::setupUi(this);
	setMinimumSize(60, 160);
	setWindowTitle("Film ");
	readSettings();				// Get remembered settings

	// Timer stuff
	load=0;						// Initial utilization
	cpuDone=false;				// Measure clock-cycles
	timer=new QTimer(this);

	// Connections
	connect(xOff, SIGNAL(valueChanged(int)), this, SLOT(xSlot(int)));
	connect(yOff, SIGNAL(valueChanged(int)), this, SLOT(ySlot(int)));
	connect(zOff, SIGNAL(valueChanged(int)), this, SLOT(zSlot(int)));

	connect(locXYZ, SIGNAL(clicked()), this, SLOT(aniXYZLocked()));
	connect(fpsBox, SIGNAL(valueChanged(int)), this, SLOT(fpsSlot(int)));

	// Call for OpenGL-update
	connect(this, SIGNAL(gotNewAni()), oglWid, SLOT(updateGLonly()));
}

// Standard destructor
aniWidget::~aniWidget() {}



/* *************************************************** */
/* GO                                                  */
/* *************************************************** */
void aniWidget::workSlot() {
	const int upTo=10;
	static int n=1, avg=0;

	sta.rotX += ((float)ani.xFac/10.0f) * offset;
	sta.rotY += ((float)ani.yFac/10.0f) * offset;
	sta.rotZ += ((float)ani.zFac/10.0f) * offset;
	emit gotNewAni();

	load=(int)(100.0*(ani.fps*secs));
	avg+=load;

	// Is averaging load-bar senseful?
	if(ani.fps<20) {
		if(load>100)
			load=100;
		loadBar->setValue(load);
	}
	else
		if(n++==upTo) {
			load=avg/upTo;
			if(load>100)
				load=100;
			loadBar->setValue(load);
			n=1;
			avg=0;
		}
}

// Start/stop animation
void aniWidget::animate(bool mode) {
	if(mode)
		timer->start(1000/ani.fps);
	else {
		timer->stop();
		loadBar->setValue(0);
	}
	emit updMaiWid();
}

// Change frames per second
void aniWidget::fpsSlot(int fps) {
	ani.fps=fps;
	if(ani.active)
		timer->start(1000/ani.fps);
}

// Start clock-cycle-measuring; called from main() once only
float aniWidget::getClock() {
	unsigned long long tmp;

	timer->setSingleShot(true);
	connect(timer, SIGNAL(timeout()), this, SLOT(cpuSlot()));

	tmp=cpuTicks();
	timer->start(220);

	while(!cpuDone)
		QApplication::processEvents();

	tmp=(cpuTicks()-tmp);

	// Reset timer and connection
	timer->stop();
	timer->setSingleShot(false);
	connect(timer, SIGNAL(timeout()), this, SLOT(workSlot()));

	return (tmp/220000.0f);
}


/* *************************************************** */
/* SET OFFSETS                                         */
/* *************************************************** */
void aniWidget::xSlot(int xVal) {
	ani.xFac=xVal;
	if(ani.locXYZ) {
		ani.yFac=ani.zFac=ani.xFac;
		updAniWid();
	}
}

void aniWidget::ySlot(int yVal) {
	ani.yFac=yVal;
	if(ani.locXYZ) {
		ani.xFac=ani.zFac=ani.yFac;
		updAniWid();
	}
}

void aniWidget::zSlot(int zVal) {
	ani.zFac=zVal;
	if(ani.locXYZ) {
		ani.xFac=ani.yFac=ani.zFac;
		updAniWid();
	}
}


/* *************************************************** */
/* UPDATE SLIDER-VALUES IN ANIMATION-EDITOR            */
/* *************************************************** */
void aniWidget::updAniWid() {
	xOff->setValue(ani.xFac);
	yOff->setValue(ani.yFac);
	zOff->setValue(ani.zFac);
	locXYZ->setChecked(ani.locXYZ);
	fpsBox->setValue(ani.fps);

	if(ani.active)
		loadBar->setValue(load);
	else
		loadBar->setValue(0);
}


/* *************************************************** */
/* HANDLE SAVE/RESTORE OF WINDOW-COORDINATES           */
/* *************************************************** */
void aniWidget::closeEvent(QCloseEvent *event) {
	writeSettings();
	event->accept();
}

void aniWidget::readSettings() {
	QSettings settings("Heinz", NAME);
	QPoint pos=settings.value("AniWindowPos", QPoint(200, 200)).toPoint();
	QSize size=settings.value("AniWindowSize", QSize(400, 400)).toSize();

	// Ensure to start with defaults on 1st program-start ever
	everStarted=false;
	everStarted=settings.value("FirstAniEdiStart").toBool();
	if(everStarted==true) {
		resize(size);
		move(pos);
	}
	else {
		everStarted=true;
		resize(QSize(165, 280));
		move(QPoint(100, 330));
	}
}

void aniWidget::writeSettings() {
	QSettings settings("Heinz", NAME);
	settings.setValue("AniWindowPos", pos());
	settings.setValue("AniWindowSize", size());
	settings.setValue("FirstAniEdiStart", everStarted);
}
