/****************************************************************************
**
** Copyright (C) 1991-2006 Heinz van Saanen
**
** This file is part of the the function viewer Zhu3D.
**
** This file may be used under the terms of the GNU General Public
** License version 2.0 as published by the Free Software Foundation
** and appearing in the file LICENSE.GPL included in the packaging of
** this file.
**
** This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
** WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
**
****************************************************************************/

#include <stdlib.h>		// For Quicksort only
#include <math.h>
#include "gldraw.h"
#include "parser.h"

// Make GUI more responsive during list generatiion
#define PE QApplication::processEvents()


// Adressing the 8 lights and error-messages
// ---------------------------------------------------------------------------
extern int lightBase;		// Global from OGLWidget
char parseResult[200];		// String for parser messages


// Some stuff for descending sorting the alpha-channels
// ---------------------------------------------------------------------------
#define MAX(a,b) (((a)>(b))?(a):(b))

typedef struct {
	int index;			// The original index
	bool parflag;		// Was an entry for paramteric system?
	GLfloat val;		// Material property for diffuse alpha-channel
} ALPHA;

static ALPHA alpha[MAXFUN+1];

// The compare-function for Quicksort
static int cmpAlpha(const void *elem1, const void *elem2) {
	if( ((ALPHA *)elem1)->val == ((ALPHA *)elem2)->val )
		return 0;
	else
		if( ((ALPHA *)elem1)->val < ((ALPHA *)elem2)->val )
			return 1;
		else
			return -1;
}

// Fetch all alpha-channel-values
static void getAlpha(void) {
	for(int i=0; i<MAXFUN+1; i++) {
		alpha[i].index=i;
		alpha[i].parflag=false;
		if(mat[i].twoSid==false)
			alpha[i].val = mat[i].difMat[0][3];
		else
			alpha[i].val = MAX(mat[i].difMat[0][3], mat[i].difMat[1][3]);
	}

	// Mark the paramtric system at last
	alpha[MAXFUN].parflag=true;
}


// Generate OpenGL-lists
// ---------------------------------------------------------------------------
static void GenFunPoints(const int n);
static void GenParPoints();
static void GenFunQuads(const int n);
static void GenParQuads();
static void GenFunTriangles(const int n);
static void GenParTriangles();


// List-Handling
// ---------------------------------------------------------------------------
bool listUp[MAXFUN+1] = {true, false, false, false};
GLuint oglLis[MAXFUN+1] = {1, 2, 3, 4};


// General
// ---------------------------------------------------------------------------
static float par0(const float xpar, const float ypar);
static float par1(const float xpar, const float ypar);
static float par2(const float xpar, const float ypar);
static void setMat(const int n);	// 0..2 for functions; 3 for parameter-mode


// Array of function-pointers used in list-generation
// ---------------------------------------------------------------------------
static float (*funPtr[MAXFUN])(float xpar, float ypar)= {fxy0, fxy1, fxy2};


// Some local comfort functions
// ---------------------------------------------------------------------------
static void normCrossProd(const float v1[3], const float v2[3], float out[3]);
static void Draw_Arrows();


/* ************************************************************************* */
/* IMPLEMENTATION PART                                                       */
/* ************************************************************************* */

// Calculating parser-values for the functions/parameter-system
// ---------------------------------------------------------------------------
float fxy0(const float xpar, const float ypar) {
	x=xpar*sta.sx;
	y=ypar*sta.sy;
	return (parse(fun[0].str)*sta.sz);
}

float fxy1(const float xpar, const float ypar) {
	x=xpar*sta.sx;
	y=ypar*sta.sy;
	return (parse(fun[1].str)*sta.sz);
}

float fxy2(const float xpar, const float ypar) {
	x=xpar*sta.sx;
	y=ypar*sta.sy;
	return (parse(fun[2].str)*sta.sz);
}

float par0(const float xpar, const float ypar) {
	s=xpar*sta.spx;
	t=ypar*sta.spy;
	return (parse(par.str[0])*sta.spz);
}

float par1(const float xpar, const float ypar) {
	s=xpar*sta.spx;
	t=ypar*sta.spy;
	return (parse(par.str[1])*sta.spz);
}

float par2(const float xpar, const float ypar) {
	s=xpar*sta.spx;
	t=ypar*sta.spy;
	return (parse(par.str[2])*sta.spz);
}


/* ************************************************************************* */
/* Starting OpenGL-Stuff                                                     */
/* ************************************************************************* */

// Generate and normalize the normal-vector
static void normCrossProd(const float v1[3], const float v2[3], float out[3]) {
	float len;

	// Calculate normal vector
	out[0]=v1[1]*v2[2] - v1[2]*v2[1];
	out[1]=v1[2]*v2[0] - v1[0]*v2[2];
	out[2]=v1[0]*v2[1] - v1[1]*v2[0];

	// Normalize it
	len=sqrt(out[0]*out[0] + out[1]*out[1] + out[2]*out[2]);
	out[0]/=len;
	out[2]/=len;
	out[3]/=len;
}

// Draws Points
// ---------------------------------------------------------------------------
void Draw_Points(void) {

	// Consider depth-buffers for transparent views
	// Get data & sort on descending alpha-channels
	getAlpha();
	qsort(alpha, MAXFUN+1, sizeof(ALPHA), &cmpAlpha);

	for(int i=0; i<MAXFUN+1; i++)
		// Check for originally parametric system
		if(alpha[i].parflag==true) {
			// Update list and view only, when needed
			if(par.drawPar) {
				setMat(alpha[i].index);
				if(listUp[alpha[i].index])
					GenParPoints();
				glCallList(oglLis[alpha[i].index]);
			}
		}
		// It was a function originally
		else {
			// Update list and view only, when needed
			if(fun[alpha[i].index].drawFun) {
				setMat(alpha[i].index);
				if(listUp[alpha[i].index])
					GenFunPoints(alpha[i].index);
				glCallList(oglLis[alpha[i].index]);
			}
		}
}

// Generate point-list
static void GenFunPoints(const int n) {
	float x, y;
	float p[4][3], v1[3], v2[3], norm[3];

	glNewList(oglLis[n], GL_COMPILE);
	glBegin(GL_POINTS);
	for(x=-1.0f; x<=1.0f; x+=sta.tessWidth)
  		for(y=-1.0; y<=1.0; y+=sta.tessWidth*2.0f) {

			p[0][0]=x;
			p[0][1]=y;
			p[0][2]=(*funPtr[n])(x, y);

			p[1][0]=x+sta.tessWidth;
			p[1][1]=y;
			p[1][2]=(*funPtr[n])(x+sta.tessWidth, y);

			p[2][0]=x;
			p[2][1]=y+sta.tessWidth;
			p[2][2]=(*funPtr[n])(x, y+sta.tessWidth);

			p[3][0]=x+sta.tessWidth;
			p[3][1]=y+sta.tessWidth;
			p[3][2]=(*funPtr[n])(x+sta.tessWidth, y+sta.tessWidth);

			v1[0]=p[3][0]-p[0][0];
			v1[1]=p[3][1]-p[0][1];
			v1[2]=p[3][2]-p[0][2];

			v2[0]=p[3][0]-p[1][0];
			v2[1]=p[3][1]-p[1][1];
			v2[2]=p[3][2]-p[1][2];

			normCrossProd(v1, v2, norm);
			glNormal3fv(norm);

			if(!elevel) {
				glVertex3fv(p[0]);	// x  y
				glVertex3fv(p[1]);	// x+ y
				glVertex3fv(p[2]);	// x  y+
				glVertex3fv(p[3]);	// x+ y+
			}
		}
	glEnd();
	glEndList();
	listUp[n]=false;
	PE;
}

// Generate point-parameter list
static void GenParPoints() {
	float s, t;
	float p[4][3], v1[3], v2[3], norm[3];

	glNewList(oglLis[3], GL_COMPILE);
	glBegin(GL_POINTS);
	for(s=-1.0f; s<=1.0f; s+=sta.tessWidth)
  		for(t=-1.0; t<=1.0; t+=sta.tessWidth*2.0f) {

			p[0][0]=par0(s,t);
			p[0][1]=par1(s,t);
			p[0][2]=par2(s,t);

			p[1][0]=par0(s+sta.tessWidth, t);
			p[1][1]=par1(s+sta.tessWidth, t);
			p[1][2]=par2(s+sta.tessWidth, t);

			p[2][0]=par0(s, t+sta.tessWidth);
			p[2][1]=par1(s, t+sta.tessWidth);
			p[2][2]=par2(s, t+sta.tessWidth);

			p[3][0]=par0(s+sta.tessWidth, t+sta.tessWidth);
			p[3][1]=par1(s+sta.tessWidth, t+sta.tessWidth);
			p[3][2]=par2(s+sta.tessWidth, t+sta.tessWidth);

			v1[0]=p[3][0]-p[0][0];
			v1[1]=p[3][1]-p[0][1];
			v1[2]=p[3][2]-p[0][2];

			v2[0]=p[3][0]-p[1][0];
			v2[1]=p[3][1]-p[1][1];
			v2[2]=p[3][2]-p[1][2];

			normCrossProd(v1, v2, norm);
			glNormal3fv(norm);

			if(!elevel) {
				glVertex3fv(p[0]);	// s  t
				glVertex3fv(p[1]);	// s+ t
				glVertex3fv(p[2]);	// s  t+
				glVertex3fv(p[3]);	// s+ t+
			}
		}
	glEnd();
	glEndList();
	listUp[3]=false;
	PE;
}


// Draws Quads
// ---------------------------------------------------------------------------
void Draw_Quads(void) {

	// Consider depth-buffers for transparent views
	// Get data & sort on descending alpha-channels
	getAlpha();
	qsort(alpha, MAXFUN+1, sizeof(ALPHA), &cmpAlpha);

	for(int i=0; i<MAXFUN+1; i++)
		// Check for originally parametric system
		if(alpha[i].parflag==true) {
			// Update list and view only, when needed
			if(par.drawPar) {
				setMat(alpha[i].index);
				if(listUp[alpha[i].index])
					GenParQuads();
				glCallList(oglLis[alpha[i].index]);
			}
		}
		// It was a function originally
		else {
			// Update list and view only, when needed
			if(fun[alpha[i].index].drawFun) {
				setMat(alpha[i].index);
				if(listUp[alpha[i].index])
					GenFunQuads(alpha[i].index);
				glCallList(oglLis[alpha[i].index]);
			}
		}
}

// Generate quad-list
static void GenFunQuads(const int n) {
	float x, y;
	float p[4][3], v1[3], v2[3], norm[3];

	glNewList(oglLis[n], GL_COMPILE);
	for(x=-1.0; x<=1.0; x+=sta.tessWidth) {
		glBegin(GL_QUAD_STRIP);
		for(y=-1.0; y<=1.0; y+=sta.tessWidth*2.0f) {

			p[0][0]=x;
			p[0][1]=y;
			p[0][2]=(*funPtr[n])(x, y);

			p[1][0]=x+sta.tessWidth;
			p[1][1]=y;
			p[1][2]=(*funPtr[n])(x+sta.tessWidth, y);

			p[2][0]=x;
			p[2][1]=y+sta.tessWidth;
			p[2][2]=(*funPtr[n])(x, y+sta.tessWidth);

			p[3][0]=x+sta.tessWidth;
			p[3][1]=y+sta.tessWidth;
			p[3][2]=(*funPtr[n])(x+sta.tessWidth, y+sta.tessWidth);

			v1[0]=p[3][0]-p[0][0];
			v1[1]=p[3][1]-p[0][1];
			v1[2]=p[3][2]-p[0][2];

			v2[0]=p[3][0]-p[1][0];
			v2[1]=p[3][1]-p[1][1];
			v2[2]=p[3][2]-p[1][2];

			normCrossProd(v1, v2, norm);
			glNormal3fv(norm);

			if(!elevel) {
				glVertex3fv(p[0]);	// x  y
				glVertex3fv(p[1]);	// x+ y
				glVertex3fv(p[2]);	// x  y+
				glVertex3fv(p[3]);	// x+ y+
			}
		}
		glEnd();
	}
	glEndList();
	listUp[n]=false;
	PE;
}

// Generate quad-parameter list
static void GenParQuads() {
	float s, t;
	float p[4][3], v1[3], v2[3], norm[3];

	glNewList(oglLis[3], GL_COMPILE);
	for(s=-1.0f; s<=1.0f; s+=sta.tessWidth) {
		glBegin(GL_QUAD_STRIP);
		for(t=-1.0f; t<=1.0f; t+=sta.tessWidth*2.0f) {

			p[0][0]=par0(s,t);
			p[0][1]=par1(s,t);
			p[0][2]=par2(s,t);

			p[1][0]=par0(s+sta.tessWidth, t);
			p[1][1]=par1(s+sta.tessWidth, t);
			p[1][2]=par2(s+sta.tessWidth, t);

			p[2][0]=par0(s, t+sta.tessWidth);
			p[2][1]=par1(s, t+sta.tessWidth);
			p[2][2]=par2(s, t+sta.tessWidth);

			p[3][0]=par0(s+sta.tessWidth, t+sta.tessWidth);
			p[3][1]=par1(s+sta.tessWidth, t+sta.tessWidth);
			p[3][2]=par2(s+sta.tessWidth, t+sta.tessWidth);

			v1[0]=p[3][0]-p[0][0];
			v1[1]=p[3][1]-p[0][1];
			v1[2]=p[3][2]-p[0][2];

			v2[0]=p[3][0]-p[1][0];
			v2[1]=p[3][1]-p[1][1];
			v2[2]=p[3][2]-p[1][2];

			normCrossProd(v1, v2, norm);
			glNormal3fv(norm);

			if(!elevel) {
				glVertex3fv(p[0]);	// s  t
				glVertex3fv(p[1]);	// s+ t
				glVertex3fv(p[2]);	// s  t+
				glVertex3fv(p[3]);	// s+ t+
			}
		}
		glEnd();
	}
	glEndList();
	listUp[3]=false;
	PE;
}


//Draws Triangles
// ---------------------------------------------------------------------------
void Draw_Triangles(void) {

	// Consider depth-buffers for transparent views
	// Get data & sort on descending alpha-channels
	getAlpha();
	qsort(alpha, MAXFUN+1, sizeof(ALPHA), &cmpAlpha);

	for(int i=0; i<MAXFUN+1; i++)
		// Check for originally parametric system
		if(alpha[i].parflag==true) {
			// Update list and view only, when needed
			if(par.drawPar) {
				setMat(alpha[i].index);
				if(listUp[alpha[i].index])
					GenParTriangles();
				glCallList(oglLis[alpha[i].index]);
			}
		}
		// It was a function originally
		else {
			// Update list and view only, when needed
			if(fun[alpha[i].index].drawFun) {
				setMat(alpha[i].index);
				if(listUp[alpha[i].index])
					GenFunTriangles(alpha[i].index);
				glCallList(oglLis[alpha[i].index]);
			}
		}
}

// Generate triangle-list
static void GenFunTriangles(const int n) {
	float x, y;
	float p[3][3], v1[3], v2[3], norm[3];

	glNewList(oglLis[n], GL_COMPILE);
	for(x=-1.0; x<=1.0; x+=sta.tessWidth) {
		glBegin(GL_TRIANGLE_STRIP);
		for(y=-1.0; y<=1.0; y+=sta.tessWidth) {

			p[0][0]=x;
			p[0][1]=y;
			p[0][2]=(*funPtr[n])(x, y);

			p[1][0]=x+sta.tessWidth;
			p[1][1]=y;
			p[1][2]=(*funPtr[n])(x+sta.tessWidth, y);

			p[2][0]=x;
			p[2][1]=y+sta.tessWidth;
			p[2][2]=(*funPtr[n])(x, y+sta.tessWidth);

			v1[0]=p[1][0]-p[0][0];
			v1[1]=p[1][1]-p[0][1];
			v1[2]=p[1][2]-p[0][2];

			v2[0]=p[2][0]-p[0][0];
			v2[1]=p[2][1]-p[0][1];
			v2[2]=p[2][2]-p[0][2];

			normCrossProd(v1, v2, norm);
			glNormal3fv(norm);

			if(!elevel) {	
				glVertex3fv(p[0]);
				glVertex3fv(p[1]);
			}
		}
		glEnd();
	}
	glEndList();
	listUp[n]=false;
	PE;
}

// Generate triangle-parameter list
static void GenParTriangles() {
	float s, t;
	float p[3][3], v1[3], v2[3], norm[3];

	glNewList(oglLis[3], GL_COMPILE);
	for(s=-1.0f; s<=1.0f; s+=sta.tessWidth) {
		glBegin(GL_TRIANGLE_STRIP);
		for(t=-1.0f; t<=1.0f; t+=sta.tessWidth) {
			p[0][0]=par0(s,t);
			p[0][1]=par1(s,t);
			p[0][2]=par2(s,t);

			p[1][0]=par0(s+sta.tessWidth, t);
			p[1][1]=par1(s+sta.tessWidth, t);
			p[1][2]=par2(s+sta.tessWidth, t);

			p[2][0]=par0(s, t+sta.tessWidth);
			p[2][1]=par1(s, t+sta.tessWidth);
			p[2][2]=par2(s, t+sta.tessWidth);

			v1[0]=p[1][0]-p[0][0];
			v1[1]=p[1][1]-p[0][1];
			v1[2]=p[1][2]-p[0][2];

			v2[0]=p[2][0]-p[0][0];
			v2[1]=p[2][1]-p[0][1];
			v2[2]=p[2][2]-p[0][2];

			normCrossProd(v1, v2, norm);
			glNormal3fv(norm);

			if(!elevel) {
				glVertex3fv(p[0]);
				glVertex3fv(p[1]);
			}
		}
		glEnd();
	}
	glEndList();
	listUp[3]=false;
	PE;
}


// Draws axes-arrows
// ---------------------------------------------------------------------------
static void Draw_Arrows() {
	const GLfloat delta=0.025f;

	// Fill triangles in any case
	if(sta.wired==true)
		glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);

	glDisable(GL_LIGHTING);
	glBegin(GL_TRIANGLES);
		glVertex3f(1.5f, -delta, 0.0f);
		glVertex3f(1.5f,  delta, 0.0f);
		glVertex3f(1.5f+2.0f*delta, 0.0f, 0.0f);

		glVertex3f(-delta, 1.5f, 0.0f);
		glVertex3f( delta, 1.5f, 0.0f);
		glVertex3f(  0.0f, 1.5f+2.0f*delta, 0.0f);

		glVertex3f( -delta, 0.0f, 1.5f);
		glVertex3f(  delta, 0.0f, 1.5f);
		glVertex3f(   0.0f, 0.0f, 1.5f+2.0f*delta);
	glEnd();
	glEnable(GL_LIGHTING);

	// Restore fill-mode if neccessary
	if(sta.wired==true)
		glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
}


// Draws axes
// ---------------------------------------------------------------------------
void Draw_Axes(void) {
	glColor3f(1.0f,1.0f,1.0f);		// Set color to white

	glDisable(GL_LIGHTING);
	glBegin(GL_LINES);
		glVertex3f( 0.0f, 0.0f, 0.0f);
		glVertex3f( 1.5f, 0.0f, 0.0f);

		glVertex3f( 0.0f, 0.0f, 0.0f);
		glVertex3f( 0.0f, 1.5f, 0.0f);

		glVertex3f( 0.0f, 0.0f, 0.0f);
		glVertex3f( 0.0f, 0.0f, 1.5f);
	glEnd();

	Draw_Arrows();
	glEnable(GL_LIGHTING);
}


// Draws Measuring-Cross
// ---------------------------------------------------------------------------
void Draw_Cross(void) {
	const GLfloat delta=0.125f;		// Cross-size

	glDisable(GL_LIGHTING);
	glColor3f(1.0f, 1.0f, 1.0f);	// Set color white
	glBegin(GL_LINES);
		glVertex2f(sta.crossX-delta,  sta.crossY);
		glVertex2f(sta.crossX+delta,  sta.crossY);
		glVertex2f(sta.crossX,        sta.crossY-delta);
		glVertex2f(sta.crossX,        sta.crossY+delta);
		glVertex3f(sta.crossX,        sta.crossY, -2.0f);
		glVertex3f(sta.crossX,        sta.crossY,  2.0f);
	glEnd();
	glEnable(GL_LIGHTING);
}


// Set material-properties
// ---------------------------------------------------------------------------
static void setMat(const int n) {

	// Mode is 1-sided
	if(ent.modOne==0.0f) {
		glMaterialfv(GL_FRONT, GL_AMBIENT,    mat[n].ambMat[0]);
		glMaterialfv(GL_FRONT, GL_DIFFUSE,    mat[n].difMat[0]);
		glMaterialfv(GL_FRONT, GL_SPECULAR,   mat[n].speMat[0]);
		glMaterialfv(GL_FRONT, GL_EMISSION,   mat[n].emiMat[0]);
		glMaterialfv(GL_FRONT, GL_SHININESS, &mat[n].shiMat[0]);
		return;
	}

	// Mode is 2-sided
	glMaterialfv(GL_FRONT, GL_AMBIENT,    mat[n].ambMat[0]);
	glMaterialfv(GL_BACK,  GL_AMBIENT,    mat[n].ambMat[1]);
	glMaterialfv(GL_FRONT, GL_DIFFUSE,    mat[n].difMat[0]);
	glMaterialfv(GL_BACK,  GL_DIFFUSE,    mat[n].difMat[1]);
	glMaterialfv(GL_FRONT, GL_SPECULAR,   mat[n].speMat[0]);
	glMaterialfv(GL_BACK,  GL_SPECULAR,   mat[n].speMat[1]);
	glMaterialfv(GL_FRONT, GL_EMISSION,   mat[n].emiMat[0]);
	glMaterialfv(GL_BACK,  GL_EMISSION,   mat[n].emiMat[1]);
	glMaterialfv(GL_FRONT, GL_SHININESS, &mat[n].shiMat[0]);
	glMaterialfv(GL_BACK,  GL_SHININESS, &mat[n].shiMat[1]);
}
