/****************************************************************************
**
** Copyright (C) 1991-2006 Heinz van Saanen
**
** This file is part of the the function viewer Zhu3D.
**
** This file may be used under the terms of the GNU General Public
** License version 2.0 as published by the Free Software Foundation
** and appearing in the file LICENSE.GPL included in the packaging of
** this file.
**
** This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
** WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
**
****************************************************************************/

#include <QtGui>
#include <GL/gl.h>
#include <QFont>
#include <math.h>
#include "glwidget.h"
#include "aniedit.h"
#include "picedit.h"
#include "funedit.h"


// For the solver: random-float [-1.0,1.0]
#define RAND() ((rand()/(float)(RAND_MAX)-0.5f)*2.0f)


// Extern globals needed
extern aniWidget *aniEdi;	// Animation currently running?
extern picWidget *picEdi;	// For pic-dimensions
extern QString workDir;	// Where to save pics as default?
extern funWidget *funEdi;	// Notify, when parameter-mode switches


/* *************************************************** */
/* Globals			                                   */
/* *************************************************** */
const int lightBase=GL_LIGHT0;					// Holds GLenum for first light source
static const GLfloat noSpot[3] = {				// For stotlights etc.
	180.0f, 180.0f, 180.0f};
static QFont xyz("System", 14, QFont::Bold);	// Font for x,y,z-markings


// Actual window-dimension of the viewer window
// Needed locally for redraw and setting new angle
// Global for pic-dimension editor
int actwidth;
int actheight;


/* *************************************************** */
/* INITIALIZATIONS                                     */
/* *************************************************** */
OGLWidget::OGLWidget(QWidget *parent) : QGLWidget(parent) {

	// Basic inits
	setMinimumSize(100, 100);
	setWindowTitle("Viewer");
	readSettings();

	// Set up picture-saving
	savePicAction = new QAction(this);
	savePicAction->setShortcut(tr("Ctrl+P"));
	this->addAction(savePicAction);
	connect(savePicAction, SIGNAL(triggered()), this, SLOT(savePic()));

	//Scalefactor; normal is default
	normal=true;
	fine=coarse=false;
	scaleFac=con.normalFac;

	// Light stuff
	sta.ediLig=sta.ediMat=0;

	// Initialize:
	iniLig();		// Lights
	iniMat();		// Materials
	iniEnt();		// Entire light
	iniAni();		// Animation
	iniSta();		// Status
	iniFun();		// Functions
	iniPar();		// Parameter values
	iniPic();		// Dimensions/lock for snapshot
}


// Standard destructor
OGLWidget::~OGLWidget() {}


// Initial OpenGL-settings
void OGLWidget::initializeGL() {
	int i;

	// Some basic initialisations
	glEnable(GL_DEPTH_TEST);
	glEnable(GL_LIGHTING);
	glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
	glShadeModel(GL_SMOOTH);
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

	// Entire light-model
	glLightModelf(GL_LIGHT_MODEL_LOCAL_VIEWER, ent.modInf);
	glLightModelf(GL_LIGHT_MODEL_TWO_SIDE, ent.modOne);

	// Setup initial light sources; spotlights are off
	for(i=0; i<MAXLIG; i++) {
		glLightfv(lightBase+i, GL_POSITION, &lig[i].ligPos[0]);
		glLightfv(lightBase+i, GL_AMBIENT,  &lig[i].ambLig[0]);
		glLightfv(lightBase+i, GL_DIFFUSE,  &lig[i].difLig[0]);
		glLightfv(lightBase+i, GL_SPECULAR, &lig[i].speLig[0]);

		glLightfv(lightBase+i, GL_CONSTANT_ATTENUATION,  &lig[i].conAtt);
		glLightfv(lightBase+i, GL_LINEAR_ATTENUATION,    &lig[i].linAtt);
		glLightfv(lightBase+i, GL_QUADRATIC_ATTENUATION, &lig[i].quaAtt);

		glLightfv(lightBase+i, GL_SPOT_DIRECTION, &lig[i].spoDir[0]);
		glLightfv(lightBase+i, GL_SPOT_CUTOFF,    &noSpot[0]);
		glLightfv(lightBase+i, GL_SPOT_EXPONENT,  &lig[i].spoExp);
	}

	// Enable/disable default lights
	for(i=0; i<MAXLIG; i++)
		if(lig[i].setLig)
			glEnable(lightBase+i);
		else
			glDisable(lightBase+i);

	// Set standard background colour
	glClearColor(ent.bacLig[0], ent.bacLig[1], ent.bacLig[2], ent.bacLig[3]);

	// "Normally" the following is not needed. But this is neccessary
	// to get a correctly initialized OpenGL-context for rendering into
	// pic-files of arbitrary size-dimensions
	listUpAll;
	if(sta.draw==0)		Draw_Triangles();
	if(sta.draw==1)		Draw_Quads();
	if(sta.draw==2)		Draw_Points();
}


// Equivalent to Glut-reshape. Widht & height is passed automatically
// Resizing is neccessary too, when changing angles
void OGLWidget::resizeGL(int width, int height) {
	glViewport(0, 0, width, height);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(sta.angle, (double)width/(double)height, 3.0, 10.0);
	glMatrixMode(GL_MODELVIEW);

	// Save actual dimensions for changing angle
	actwidth=width;
	actheight=height;

	// Update pic-dimension editor at last
	picEdi->updPicWid();
}


/* *************************************************** */
/* ROTATIONS;    x = value from xSlider or mouse       */
/* *************************************************** */
void OGLWidget::setXRotation(const int newx) {
	if(newx!=sta.rotX) {
		sta.rotX=newx%361;
		emit xRotationChanged(newx);
		updateGL();
	}
}

void OGLWidget::setYRotation(const int newy) {
	if(newy!=sta.rotY) {
		sta.rotY=newy%361;
		emit yRotationChanged(newy);
		updateGL();
	}
}

void OGLWidget::setZRotation(const int newz) {
	if(newz!=sta.rotZ) {
		sta.rotZ=newz;
		updateGL();
	}
}


/* *************************************************** */
/* GOT NEW FUNCTIONS                                   */
/* *************************************************** */
void OGLWidget::setNewF0() {
	QString message="Parser is unhappy with F0/U\n\n";

	// Convert and check for correctness
	if(!sta.parMod)
		parse(fun[0].str);
	else
		parse(par.str[0]);

	if(elevel) {
		message=message+errorStr[elevel-1];
		QMessageBox::critical(this, tr("Alert"), message);
		return;
	}

	if(!sta.parMod)
		listUp[0]=true;
	else
		listUp[3]=true;

	updateGL();
	emit updMaiInf();
}

void OGLWidget::setNewF1() {
	QString message="Parser is unhappy with F1/V\n\n";

	// Convert and check for correctness
	if(!sta.parMod)
		parse(fun[1].str);
	else
		parse(par.str[1]);

	if(elevel) {
		message=message+errorStr[elevel-1];
		QMessageBox::critical(this, tr("Alert"), message);
		return;
	}

	if(!sta.parMod)
		listUp[1]=true;
	else
		listUp[3]=true;

	updateGL();
	emit updMaiInf();
}

void OGLWidget::setNewF2() {
	QString message="Parser is unhappy with F2/W\n\n";

	// Convert and check for correctness
	if(!sta.parMod)
		parse(fun[2].str);
	else
		parse(par.str[2]);

	if(elevel) {
		message=message+errorStr[elevel-1];
		QMessageBox::critical(this, tr("Alert"), message);
		return;
	}

	if(!sta.parMod)
		listUp[2]=true;
	else
		listUp[3]=true;

	updateGL();
	emit updMaiInf();
}


/* *************************************************** */
/* HANDLE POLYGON-MODE                                 */
/* *************************************************** */
void OGLWidget::togglePolygon() {
	if(sta.draw==2)	// Points? Then do nothing
		return;
	sta.wired = !sta.wired;
	updateGL();
}


/* *************************************************** */
/* PARAMETER-MODE CHANGED                              */
/* *************************************************** */
void OGLWidget::drawModeParameter() {
	sta.parMod=!sta.parMod;
	emit updMaiWid();
	funEdi->updFunWid();
}


/* *************************************************** */
/* PERFORM SCALINGS                                    */
/* *************************************************** */
// Scale x
void OGLWidget::scaleXplus()  {
	if(!sta.parMod) {
		sta.sx*=scaleFac;
		listUpFun;
		emit updMaiInf();
		updateGL();
	}
	else  {
		sta.spx*=scaleFac;
		listUpPar;
		emit updMaiInf();
		updateGL();
	}
}

void OGLWidget::scaleXminus() {
	if(!sta.parMod) {
		sta.sx/=scaleFac;
		listUpFun;
		emit updMaiInf();
		updateGL();
	}
	else {
		sta.spx/=scaleFac;
		listUpPar;
		emit updMaiInf();
		updateGL();
	}
}

// Scale y
void OGLWidget::scaleYplus()  {
	if(!sta.parMod) {
		sta.sy*=scaleFac;
		listUpFun;
		emit updMaiInf();
		updateGL();
	}
	else {
		sta.spy*=scaleFac;
		listUpPar;
		emit updMaiInf();
		updateGL();
	}
}

void OGLWidget::scaleYminus() {
	if(!sta.parMod) {
		sta.sy/=scaleFac;
		listUpFun;
		emit updMaiInf();
		updateGL();
	}
	else {
		sta.spy/=scaleFac;
		listUpPar;
		emit updMaiInf();
		updateGL();
	}
}

// Scale z
void OGLWidget::scaleZplus()  {
	if(!sta.parMod) {
		sta.sz*=scaleFac;
		listUpFun;
		emit updMaiInf();
		updateGL();
	}
	else {
		sta.spz*=scaleFac;
		listUpPar;
		emit updMaiInf();
		updateGL();
	}
}

void OGLWidget::scaleZminus() {
	if(!sta.parMod) {
		sta.sz/=scaleFac;
		listUpFun;
		emit updMaiInf();
		updateGL();
	}
	else {
		sta.spz/=scaleFac;
		listUpPar;
		emit updMaiInf();
		updateGL();
	}
}

// Scale xyz
void OGLWidget::scaleXYZplus()  {
	sta.sx=sta.sy=sta.sz=sta.sa*=scaleFac;
	if(!sta.parMod) {
		listUpFun;
		emit updMaiInf();
		updateGL();
	}
}

void OGLWidget::scaleXYZminus() {
	sta.sx=sta.sy=sta.sz=sta.sa/=scaleFac;
	if(!sta.parMod) {
		listUpFun;
		emit updMaiInf();
		updateGL();
	}
}

// Scale grid
void OGLWidget::scaleGridplus() {
	if(sta.tessWidth>0.005f) {
		sta.tessWidth/=con.tessFac;
		listUpAll;
		updateGL();
	}
}

void OGLWidget::scaleGridminus() {
	if(sta.tessWidth<0.5f) {
		sta.tessWidth*=con.tessFac;
		listUpAll;
		updateGL();
	}
}


/* *************************************************** */
/* MOUSE-HANDLING                                      */
/* *************************************************** */
// Qt is missing this. Quite a pitty:-) 
static bool LeftAndRightButtons(QMouseEvent *event) {
	if(event->buttons() & Qt::LeftButton)
		if(event->buttons() & Qt::RightButton)
			return true;
	if(event->buttons() & Qt::RightButton)
		if(event->buttons() & Qt::LeftButton)
			return true;
	return false;
}

void OGLWidget::mousePressEvent(QMouseEvent *event) {
	if(event->buttons() & LeftAndRightButtons(event))
		setCursor(Qt::SizeAllCursor);
	else if(event->buttons() & Qt::MidButton)
			setCursor(Qt::SizeVerCursor);
	else if(event->buttons() & Qt::LeftButton)
			setCursor(Qt::PointingHandCursor);
	else if(event->buttons() & Qt::RightButton)
			setCursor(Qt::CrossCursor);
	lastPos=event->pos();
}

void OGLWidget::mouseReleaseEvent(QMouseEvent *event) {
	setCursor(Qt::ArrowCursor);
	lastPos=event->pos();
}

void OGLWidget::mouseMoveEvent(QMouseEvent *event) {

	const float delta=1.0f/(float)sta.grids;
	int dx = event->x()-lastPos.x();
	int dy = event->y()-lastPos.y();

	// Zoom in/out with MidButton
	if(event->buttons() & Qt::MidButton) {
		if(dy>0)
			zoomOut();
		else
			zoomIn();
		lastPos = event->pos();
		return;
	}

	// Translate with LeftAndRightButtons
	if(event->buttons() & LeftAndRightButtons(event)) {
		if(dx>0)
			transRight();
 		else if(dx<0)
			transLeft();

		if(dy>0)
			transDown();
		else if(dy<0)
			transUp();

		lastPos = event->pos();
		return;
	}

	// Rotate view with left button
	if(event->buttons() & Qt::LeftButton) {
		setXRotation((int)sta.rotX+1*dy);
		setYRotation((int)sta.rotY+1*dx);
		updateGL();
		lastPos = event->pos();
		return;
	}

	// Move cross with right button
	if(event->buttons() & Qt::RightButton) {
		if(dx>0) {
			sta.crossX+=delta;
			if(sta.crossX>1.0f) sta.crossX=1.0f;
		}
 		else if(dx<0) {
			sta.crossX-=delta;
			if(sta.crossX<-1.0f) sta.crossX=-1.0f;
		}

		if(dy>0) {
			sta.crossY-=delta;
			if(sta.crossY<-1.0f) sta.crossY=-1.0f;
		}
		else if(dy<0) {
			sta.crossY+=delta;
			if(sta.crossY>1.0f) sta.crossY=1.0f;
		}
	}

	updateGL();
	emit updMaiInf();
	lastPos = event->pos();
}


/* *************************************************** */
/* SAVE PICTURE                                        */
/* *************************************************** */
void OGLWidget::savePic() {

	QString fileName;
	QString title, x, y;
	bool aniFlag=ani.active;	// Remember if animation was on
	int widthbak=actwidth;		// Remember actual OGL-width
	int heightbak=actheight;	// Remember actual OGL-height

	// Adjust some temporary settings
	aniEdi->animate(false);
	x.setNum(pic.xDim);
	y.setNum(pic.yDim);
	title="Rendering "+x+"x"+y+" picture ...";
	oglWid->setWindowTitle(title);

	// Render into pixmap & restore original view
	QApplication::setOverrideCursor(Qt::WaitCursor);
	QPixmap pixmap = oglWid->renderPixmap(pic.xDim, pic.yDim);
	resizeGL(widthbak, heightbak);
	QApplication::restoreOverrideCursor();

	// Restore original oglWid-window-name
	if(aniFlag==true) {
		oglWid->setWindowTitle("Animation");
		aniEdi->animate(true);
	}
	else
		oglWid->setWindowTitle("Viewer");

	// Save pixmap
	if(pic.png==true) {
		fileName=QFileDialog::getSaveFileName(
					this,
					"Save as *.png",
					workDir,
					"Images (*.png)");

		if(fileName.isEmpty())
			return;

		// Complete filename
		if(!fileName.toUpper().endsWith(".PNG"))
			fileName+=".png";

		QApplication::setOverrideCursor(Qt::WaitCursor);
		pixmap.save(fileName, "PNG", pic.quality);
		QApplication::restoreOverrideCursor();
	}
	else {
		fileName=QFileDialog::getSaveFileName(
					this,
					"Save as *.jpg",
					workDir,
					"Images (*.jpg)");

		if(fileName.isEmpty())
			return;

		// Complete filename
		if(!fileName.toUpper().endsWith(".JPG"))
			fileName+=".jpg";

		QApplication::setOverrideCursor(Qt::WaitCursor);
		pixmap.save(fileName, "JPG", pic.quality);
		QApplication::restoreOverrideCursor();
	}

	// Consequently update current workDir
	fileName.resize(fileName.lastIndexOf('/'));
	workDir=fileName;
}


/* *************************************************** */
/* MISCELLANEOUS GOODIES                               */
/* *************************************************** */

// When oglWid is killed, close MainWindow too
// Using qApp->closeAllWindows() instead freezes application !
void OGLWidget::closeEvent(QCloseEvent *event) {
	writeSettings();
	emit killAll();
	event->accept();
}

void OGLWidget::readSettings() {
	QSettings settings("Heinz", NAME);
	QPoint pos=settings.value("oglWidPos", QPoint(200, 200)).toPoint();
	QSize size=settings.value("oglWidSize", QSize(400, 400)).toSize();

	// Ensure to start with defaults on 1st program-start ever
	everStarted=false;
	everStarted=settings.value("FirstOGLStart").toBool();
	if(everStarted==true) {
		resize(size);
		move(pos);
	}
	else {
		everStarted=true;
		resize(QSize(570, 570));
		move(QPoint(445, 10));
	}
}

void OGLWidget::writeSettings() {
	QSettings settings("Heinz", NAME);
	settings.setValue("oglWidPos", pos());
	settings.setValue("oglWidSize", size());
	settings.setValue("FirstOGLStart", everStarted);
}


/* *************************************************** */
/* LIGHTNING STUFF                                     */
/* *************************************************** */

/* *************************************************** */
/* DISTANCE FROM ORIGIN                                */
/* *************************************************** */
void OGLWidget::lightDistance(const int d) {
	const float off=d/10.0f;
	switch(sta.ediLig) {
		case 0:		lig[0].ligPos[0]=-off;
					lig[0].ligPos[1]=off;
					lig[0].ligPos[2]=off;
					break;
		case 1:		lig[1].ligPos[0]=-off;
					lig[1].ligPos[1]=-off;
					lig[1].ligPos[2]=off;
					break;
		case 2:		lig[2].ligPos[0]=off;
					lig[2].ligPos[1]=-off;
					lig[2].ligPos[2]=off;
					break;
		case 3:		lig[3].ligPos[0]=off;
					lig[3].ligPos[1]=off;
					lig[3].ligPos[2]=off;
					break;
		case 4:		lig[4].ligPos[0]=-off;
					lig[4].ligPos[1]=off;
					lig[4].ligPos[2]=-off;
					break;
		case 5:		lig[5].ligPos[0]=-off;
					lig[5].ligPos[1]=-off;
					lig[5].ligPos[2]=-off;
					break;
		case 6:		lig[6].ligPos[0]=off;
					lig[6].ligPos[1]=-off;
					lig[6].ligPos[2]=-off;
					break;
		case 7:		lig[7].ligPos[0]=off;
					lig[7].ligPos[1]=off;
					lig[7].ligPos[2]=-off;
					break;
	}
	lig[sta.ediLig].ligDis=d;
	updateGL();
}


/* *************************************************** */
/* Ambient light intensity                             */
/* *************************************************** */
void OGLWidget::laall(const int channel, const int intensity) {
	if(lig[sta.ediLig].ambLigLoc && channel<3)
		for(int i=0; i<3; i++) {
				lig[sta.ediLig].ambLig[i]=intensity/255.0f;
				emit updLigAmb();
		}
	else
		lig[sta.ediLig].ambLig[channel]=intensity/255.0f;
	glLightfv(lightBase+sta.ediLig, GL_AMBIENT, &lig[sta.ediLig].ambLig[0]);
	updateGL();
}


/* *************************************************** */
/* Diffuse light intensity                             */
/* *************************************************** */
void OGLWidget::ldall(const int channel, const int intensity) {
	if(lig[sta.ediLig].difLigLoc && channel<3)
		for(int i=0; i<3; i++) {
				lig[sta.ediLig].difLig[i]=intensity/255.0f;
				emit updLigDif();
		}
	else
		lig[sta.ediLig].difLig[channel]=intensity/255.0f;
	glLightfv(lightBase+sta.ediLig, GL_DIFFUSE, &lig[sta.ediLig].difLig[0]);
	updateGL();
}


/* *************************************************** */
/* Specular light intensity                            */
/* *************************************************** */
void OGLWidget::lsall(const int channel, const int intensity) {
	if(lig[sta.ediLig].speLigLoc && channel<3)
		for(int i=0; i<3; i++) {
				lig[sta.ediLig].speLig[i]=intensity/255.0f;
				emit updLigSpe();
		}
	else
		lig[sta.ediLig].speLig[channel]=intensity/255.0f;
	glLightfv(lightBase+sta.ediLig, GL_SPECULAR, &lig[sta.ediLig].speLig[0]);
	updateGL();
}


/* *************************************************** */
/* Attenuation                                         */
/* *************************************************** */
void OGLWidget::setConAtt(int faktor) {
	lig[sta.ediLig].conAtt=faktor/100.0f;
	glLightfv(lightBase+sta.ediLig, GL_CONSTANT_ATTENUATION, &lig[sta.ediLig].conAtt);
	updateGL();
}

void OGLWidget::setLinAtt(int faktor) {
	lig[sta.ediLig].linAtt=faktor/300.0f;
	glLightfv(lightBase+sta.ediLig, GL_LINEAR_ATTENUATION, &lig[sta.ediLig].linAtt);
	updateGL();
}

void OGLWidget::setQuaAtt(int faktor) {
	lig[sta.ediLig].quaAtt=(float)faktor/1000.0f;
	glLightfv(lightBase+sta.ediLig, GL_QUADRATIC_ATTENUATION, &lig[sta.ediLig].quaAtt);
	updateGL();
}


/* *************************************************** */
/* MATERIAL STUFF                                      */
/* *************************************************** */

/* *************************************************** */
/* AMBIENT MATERIAL                                    */
/* *************************************************** */
void OGLWidget::maall(const int channel, const int intensity) {
	if(mat[sta.ediMat].ambMatLoc[mat[sta.ediMat].twoSid] && channel<3)
		for(int i=0; i<3; i++) {
				mat[sta.ediMat].ambMat[mat[sta.ediMat].twoSid][i]=intensity/255.0f;
				emit updMatWid();
		}
	else
		mat[sta.ediMat].ambMat[mat[sta.ediMat].twoSid][channel]=intensity/255.0f;
	updateGL();
}


/* *************************************************** */
/* DIFFUSE MATERIAL                                    */
/* *************************************************** */
void OGLWidget::mdall(const int channel, const int intensity) {
	if(mat[sta.ediMat].difMatLoc[mat[sta.ediMat].twoSid] && channel<3)
		for(int i=0; i<3; i++) {
				mat[sta.ediMat].difMat[mat[sta.ediMat].twoSid][i]=intensity/255.0f;
				emit updMatWid();
		}
	else
		mat[sta.ediMat].difMat[mat[sta.ediMat].twoSid][channel]=intensity/255.0f;
	updateGL();
}


/* *************************************************** */
/* SPECULAR MATERIAL                                   */
/* *************************************************** */
void OGLWidget::msall(const int channel, const int intensity) {
	if(mat[sta.ediMat].speMatLoc[mat[sta.ediMat].twoSid] && channel<3)
		for(int i=0; i<3; i++) {
				mat[sta.ediMat].speMat[mat[sta.ediMat].twoSid][i]=intensity/255.0f;
				emit updMatWid();
		}
	else
		mat[sta.ediMat].speMat[mat[sta.ediMat].twoSid][channel]=intensity/255.0f;
	updateGL();
}


/* *************************************************** */
/* EMISSION MATERIAL                                   */
/* *************************************************** */
void OGLWidget::meall(const int channel, const int intensity) {
	if(mat[sta.ediMat].emiMatLoc[mat[sta.ediMat].twoSid] && channel<3)
		for(int i=0; i<3; i++) {
				mat[sta.ediMat].emiMat[mat[sta.ediMat].twoSid][i]=intensity/255.0f;
				emit updMatWid();
		}
	else
		mat[sta.ediMat].emiMat[mat[sta.ediMat].twoSid][channel]=intensity/255.0f;
	updateGL();
}


/* *************************************************** */
/* TOGGLE MATERIAL EDITED ACTUALLY                     */
/* *************************************************** */
void OGLWidget::toggleBackSide() {

	if(ent.modOne==0.0f && !mat[sta.ediMat].twoSid) {
		QMessageBox::information(this,
					tr("Material"),
					tr(	"Editing back-side colours, although the "
						"global setting is just 1-sided.<p></p>"
						"Zhu3D is <b>enabling</b> the 2-sided mode automatically now."));
		ent.modOne=1.0f;
		updEntWid();
	}

	mat[sta.ediMat].twoSid = !mat[sta.ediMat].twoSid;
	updMatWid();
}


/* *************************************************** */
/* SYNCRONIZE BACKSIDE WITH FRONTSIDE                  */
/* *************************************************** */
void OGLWidget::syncSides() {
	int j;

	if(ent.modOne==0.0f) {
		QMessageBox::information(this,
					tr("Material"),
					tr(	"Changing back-side colours, although the "
						"global setting is just 1-sided.<p></p>"
						"Zhu3D is <b>enabling</b> the 2-sided mode automatically now."));
		ent.modOne=1.0f;
		updEntWid();
	}

	mat[sta.ediMat].ambMatLoc[1]=mat[sta.ediMat].ambMatLoc[0];
	mat[sta.ediMat].difMatLoc[1]=mat[sta.ediMat].difMatLoc[0];
	mat[sta.ediMat].speMatLoc[1]=mat[sta.ediMat].speMatLoc[0];
	mat[sta.ediMat].emiMatLoc[1]=mat[sta.ediMat].emiMatLoc[0];
	for(j=0; j<4; j++) mat[sta.ediMat].ambMat[1][j]=mat[sta.ediMat].ambMat[0][j];
	for(j=0; j<4; j++) mat[sta.ediMat].difMat[1][j]=mat[sta.ediMat].difMat[0][j];
	for(j=0; j<4; j++) mat[sta.ediMat].speMat[1][j]=mat[sta.ediMat].speMat[0][j];
	for(j=0; j<4; j++) mat[sta.ediMat].emiMat[1][j]=mat[sta.ediMat].emiMat[0][j];
	mat[sta.ediMat].shiMat[1]=mat[sta.ediMat].shiMat[0];

	updMatWid();
	updateGL();
}


/* *************************************************** */
/* ENTIRE STUFF                                        */
/* *************************************************** */

/* *************************************************** */
/* Entire background colour                            */
/* *************************************************** */
void OGLWidget::lmball(const int channel, const int intensity) {
	if(ent.bacEntLoc && channel<3)
		for(int i=0; i<3; i++) {
				ent.bacLig[i]=intensity/255.0f;
				emit updEntWid();
		}
	else
		ent.bacLig[channel]=intensity/255.0f;
	updateGL();
}


/* *************************************************** */
/* Entire ambient light modell                         */
/* *************************************************** */
void OGLWidget::lmaall(const int channel, const int intensity) {
	if(ent.ambEntLoc && channel<3)
		for(int i=0; i<3; i++) {
				ent.modLig[i]=intensity/255.0f;
				emit updEntWid();
		}
	else
		ent.modLig[channel]=intensity/255.0f;
	updateGL();
}


/* ************************************************************************ */
/* paintGL() and solve()                                                    */
/* ************************************************************************ */

/* *************************************************** */
/* Solve via adaptive random-search                    */
/* *************************************************** */
void OGLWidget::solve() {

	// Central constants
	const int maxsteps=75000;
	const float eps=5e-7f;
 
	QString message="Solver is unhappy with ";
	char tmpString[100];
	int i=0;
	float x, y, z, dx, dy, hz;
	bool ready=false;

	x=sta.crossX;
	y=sta.crossY;
	sprintf(parseResult, "Start solving at: X=%.6f  Y=%.6f\n",
			x*sta.sx, y*sta.sy);

	// Parse the three functions and return on failure
	parse(fun[0].str); 
	if(elevel) {
		message=message+"Function 0          \n\n"+errorStr[elevel-1];
		QMessageBox::critical(this, tr("Alert"), message);
		emit parserReady();
		return;
	}

	parse(fun[1].str); 
	if(elevel) {
		message=message+"Function 1          \n\n"+errorStr[elevel-1];
		QMessageBox::critical(this, tr("Alert"), message);
		emit parserReady();
		return;
	}

	parse(fun[2].str); 
	if(elevel) {
		message=message+"Function 2          \n\n"+errorStr[elevel-1];
		QMessageBox::critical(this, tr("Alert"), message);
		emit parserReady();
		return;
	}

	// Calculate the difference of the Z-values from X,Y-cross-value
	QApplication::setOverrideCursor(Qt::WaitCursor);
	TimerStart();
	z =fabs(fxy1(x, y)-fxy0(x, y));
	z+=fabs(fxy2(x, y)-fxy0(x, y));
	z+=fabs(fxy2(x, y)-fxy1(x, y));

	do {
		dx=RAND()*eps;		// Generate a new random-vector
		dy=RAND()*eps;		// with length eps

		// Calculate Z-differences to it
		hz =fabs(fxy1(x+dx, y+dy)-fxy0(x+dx, y+dy));
		hz+=fabs(fxy2(x+dx, y+dy)-fxy0(x+dx, y+dy));
		hz+=fabs(fxy2(x+dx, y+dy)-fxy1(x+dx, y+dy));

		while(hz<z) {
			if(++i>maxsteps)
				break;
			if(hz<eps) {
				ready=true;
				break;
			}
			z=hz;
			dx*=2.0f;	// Resize random-vector dynamically
			dy*=2.0f;
			x+=dx;
			y+=dy;

			hz =fabs(fxy1(x+dx, y+dy)-fxy0(x+dx, y+dy));
			hz+=fabs(fxy2(x+dx, y+dy)-fxy0(x+dx, y+dy));
			hz+=fabs(fxy2(x+dx, y+dy)-fxy1(x+dx, y+dy));
		}

		if(ready==true)
			break;

		// Last random-search wasn't successfull; Take back one step
		x-=dx;
		y-=dy;
		z =fabs(fxy1(x, y)-fxy0(x, y));
		z+=fabs(fxy2(x, y)-fxy0(x, y));
		z+=fabs(fxy2(x, y)-fxy1(x, y));
	} while (i++<maxsteps);

	// Generate info-strings for mainWindow
	TimerStop();
	QApplication::restoreOverrideCursor();
	sprintf(tmpString, "Calculated %d iterations in %.2f mSec\nResult is %srealiable.\n",
			i-1, secs*1000.0f,  (ready==true)?"":"NOT ");
	strcat(parseResult, tmpString);
	sprintf(tmpString, "\nSolution: X=%.6f  Y=%.6f  Z=%.6f",
			x*sta.sx, y*sta.sy, fxy0(x,y)/sta.sz);
	strcat(parseResult, tmpString);

	// Alert on unreliable solution
	if(ready==false)
		QMessageBox::critical(this,
								tr("Solver"),
								tr("Result is NOT reliable"));
	// Set cross to reliable coordinates;
	else {
		sta.crossX=x;
		sta.crossY=y;
	}

	// Deliver infos to mainWindow
	emit parserReady();
	emit updMaiInf();
	updateGL();
} // End of solve()


// Central Drawing-Routine for GL-viewer
// -----------------------------------------------------------------
void OGLWidget::paintGL() {

	// Delete framebuffer && Z-buffer
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	// Initialize modelview-matrix
	glLoadIdentity();

	// Set lights
	for(int i=0; i<MAXLIG; i++) {

		// Sitch lights on/off
		if(lig[i].setLig)
			glEnable(lightBase+i);
		else
			glDisable(lightBase+i);

		// Set light-positions with inf./local
		lig[i].ligPos[3]=ent.modInf;
		glLightfv(lightBase+i, GL_POSITION, &lig[i].ligPos[0]);
		
		// Set light-shares
		glLightfv(lightBase+i, GL_AMBIENT,  &lig[i].ambLig[0]);
		glLightfv(lightBase+i, GL_DIFFUSE,  &lig[i].difLig[0]);
		glLightfv(lightBase+i, GL_SPECULAR, &lig[i].speLig[0]);

		// Set attenution-factors
		glLightfv(lightBase+i, GL_CONSTANT_ATTENUATION,  &lig[i].conAtt);
		glLightfv(lightBase+i, GL_LINEAR_ATTENUATION,    &lig[i].linAtt);
		glLightfv(lightBase+i, GL_QUADRATIC_ATTENUATION, &lig[i].quaAtt);

		// Set spotlights
		if(lig[i].spoSta==false)
			glLightfv(lightBase+i, GL_SPOT_CUTOFF, noSpot);
		else
			glLightfv(lightBase+i, GL_SPOT_CUTOFF, &lig[i].spoCut);
		glLightfv(lightBase+i, GL_SPOT_EXPONENT, &lig[i].spoExp);
		glLightfv(lightBase+i, GL_SPOT_DIRECTION, &lig[i].spoDir[0]);

	}

	// Entire part
	glLightModelf(GL_LIGHT_MODEL_LOCAL_VIEWER, ent.modInf); 
	glLightModelf(GL_LIGHT_MODEL_TWO_SIDE, ent.modOne);
	glLightModelfv(GL_LIGHT_MODEL_AMBIENT, &ent.modLig[0]);
	glClearColor(ent.bacLig[0], ent.bacLig[1], ent.bacLig[2], ent.bacLig[3]);

	// Do model-transformations
	glTranslatef(0.0, 0.0 , sta.transZ);  			// Translate in Z-direction
	glRotatef(sta.rotX, 1.0, 0.0, 0.0);  			// Rotate around X-axis
	glRotatef(sta.rotY, 0.0, 1.0, 0.0);  			// Rotate around Y-axis
	glRotatef(sta.rotZ, 0.0, 0.0, 1.0);  			// Rotate around Z-axis
	glTranslatef(sta.transX, sta.transY , -0.5); 	// Translate all with -0.5

	// Check polygon-mode
	if(sta.wired==true)
		glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
	else
		glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);

	// And draw under current model-transformation finally
	if(sta.draw==0)			Draw_Triangles();
	if(sta.draw==1)			Draw_Quads();
	if(sta.draw==2)			Draw_Points();
	if(sta.cross==true)	Draw_Cross();
	if(sta.axes==true) {
		glDisable(GL_LIGHTING);
		qglColor(QColor(Qt::white));
		xyz.setStyleStrategy(QFont::OpenGLCompatible);
		renderText( 1.5, -0.1, 0.0, "x", xyz);
		renderText( 0.1,  1.5, 0.0, "y", xyz);
		renderText(-0.1, -0.1, 1.5, "z", xyz);
		glEnable(GL_LIGHTING);
		Draw_Axes();
	}
} // End of paintGL()
