/* ******************************************************************************
**
** Copyright (C) 1991-2006 Heinz van Saanen
**
** This file is part of the the function viewer Zhu3D.
**
** This file may be used under the terms of the GNU General Public
** License version 2.0 as published by the Free Software Foundation
** and appearing in the file LICENSE.GPL included in the packaging of
** this file.
**
** This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
** WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
**
********************************************************************************** */

#include <QtGui>
#include "mainwindow.h"
#include "timers.h"


// Consider eventual arguments for qmake
#ifndef DOCDIR
	#define MYDOCDIR (appPath+"/doc/")
#else
	#define MYDOCDIR (QString(DOCDIR)+"/")
#endif

#ifndef SYSDIR
	#define MYSYSDIR (appPath+"/system/")
#else
	#define MYSYSDIR (QString(SYSDIR)+"/")
#endif


// The current workDir will be remembered after after application-exit
// Global, because it is needed from oglWidget for saving pics
QString workDir;

// When 0.0f after main(), timestamp is not implemented for current CPU
// readSettings() gives a message during 1.st start then
extern float MHz;


/* ********************************************************************** */
/* MAINWINDOW CLASS                                                       */
/* ********************************************************************** */
MainWindow::MainWindow() {

	// Set up the main control-center
	controlCenter = new QWidget;
	Ui_mainCtrl::setupUi(controlCenter);
	this->setMinimumSize(300, 350);
	setWindowTitle(NAME);
	setCentralWidget(controlCenter);

	// Initialize settings
	const QString startFileName=MYSYSDIR+QString::fromAscii("startup.zhu", 11);
	workDir=appPath+"/work/";

	// Create help- and solver-widget
	hlpEdi = new QTextEdit();
	hlpEdi->setReadOnly(true);
	solEdi = new QTextEdit();
	solEdi->setWindowTitle("Solver report");
	solEdi->setMinimumSize(250, 150);

	// And update them, if something was remembered
	// Considered are settings like size, pos, workDir, ...
	readSettings();

	// When debugging, show extra error-window
	#ifdef DEBUG
	errEdi->show();
	#endif

	// Create actions; Done now, because of file loading
	createActions();
	createMenus();
	createToolBars();

	// Opened new file or update GL-widget?
	// Done now, because of immediate file loading
	connect(this, SIGNAL(gotNewFile()), oglWid, SLOT(newFileOpened()));
	connect(this, SIGNAL(gotNewView()), oglWid, SLOT(updateGLonly()));

	// Was there a file as argument or exists the startup-file?
	if(gotArg)
		if(loadFile(argStr))
			emit gotNewFile();
		else
			QMessageBox::critical(this,
							tr("Program start\n"),
							tr("<b>Error</b> when loading file."
								"<p>Zhu3D proceeds with built-in "
								"standard modell.</p>")
			);
	else
		if(loadFile(startFileName))
			emit gotNewFile();
		else
			QMessageBox::information(this,
							tr("Program start\n"),
							tr("Missing <b>startup.zhu</b> in your path."
								"<p>Zhu3D proceeds with built-in "
								"standard modell.</p>")
			);

	// Initialize the rest
	fineChecked=false;			// Normal is default for scaling
	normalChecked=true;
	coarseChecked=false;
	oglWid->show();				// Show the viewer window
	updMaiWid();				// Update mainwindow-widget
	updMaiInf();				// Update mainwindow-info-messages

	//Got animation instead of "normal" file?
	if(ani.active) {
		runStopAction->setIcon(QIcon(":/images/stop.png"));
		runStopAction->setToolTip("Stop animation");
		oglWid->setWindowTitle("Animation");
		aniEdi->animate(true);
	}

	// CONNECTIONS
	// --------------------------------------------------------------------
	// Slider-signals and rotations-slots
	connect(xSlider, SIGNAL(valueChanged(int)), oglWid, SLOT(setXRotation(int)));
	connect(ySlider, SIGNAL(valueChanged(int)), oglWid, SLOT(setYRotation(int)));
	connect(zSlider, SIGNAL(valueChanged(int)), oglWid, SLOT(setZRotation(int)));

	// Mouse-signals and slider-position
	connect(oglWid, SIGNAL(xRotationChanged(int)), xSlider, SLOT(setValue(int)));
	connect(oglWid, SIGNAL(yRotationChanged(int)), ySlider, SLOT(setValue(int)));

	// Set modes
	connect(Point,     SIGNAL(clicked()), oglWid, SLOT(drawModePoint()));
	connect(Triangle,  SIGNAL(clicked()), oglWid, SLOT(drawModeTriangle()));
	connect(Quad,      SIGNAL(clicked()), oglWid, SLOT(drawModeQuad()));
	connect(Parameter, SIGNAL(clicked()), oglWid, SLOT(drawModeParameter()));
	connect(Point,     SIGNAL(clicked()), this, SLOT(wireModeSlot()));
	connect(Triangle,  SIGNAL(clicked()), this, SLOT(wireModeSlot()));
	connect(Quad,      SIGNAL(clicked()), this, SLOT(wireModeSlot()));
	connect(Parameter, SIGNAL(clicked()), this, SLOT(wireModeSlot()));

	// Enable/disable functions, cross ...
	connect(Function0, SIGNAL(clicked()), oglWid, SLOT(toggleF0()));	// Functions
	connect(Function1, SIGNAL(clicked()), oglWid, SLOT(toggleF1()));
	connect(Function2, SIGNAL(clicked()), oglWid, SLOT(toggleF2()));
	connect(Function3, SIGNAL(clicked()), oglWid, SLOT(toggleF3()));	// Parameters

	connect(Cross, SIGNAL(clicked()), oglWid, SLOT(toggleCross()));
	connect(Axes,  SIGNAL(clicked()), oglWid, SLOT(toggleAxes()));
	connect(Wired, SIGNAL(clicked()), oglWid, SLOT(togglePolygon()));

	// Handle scaling
	connect(xPlus,     SIGNAL(clicked()), oglWid, SLOT(scaleXplus()));
	connect(xMinus,    SIGNAL(clicked()), oglWid, SLOT(scaleXminus()));
	connect(yPlus,     SIGNAL(clicked()), oglWid, SLOT(scaleYplus()));
	connect(yMinus,    SIGNAL(clicked()), oglWid, SLOT(scaleYminus()));
	connect(zPlus,     SIGNAL(clicked()), oglWid, SLOT(scaleZplus()));
	connect(zMinus,    SIGNAL(clicked()), oglWid, SLOT(scaleZminus()));
	connect(xyzPlus,   SIGNAL(clicked()), oglWid, SLOT(scaleXYZplus()));
	connect(xyzMinus,  SIGNAL(clicked()), oglWid, SLOT(scaleXYZminus()));
	connect(gridPlus,  SIGNAL(clicked()), oglWid, SLOT(scaleGridplus()));
	connect(gridMinus, SIGNAL(clicked()), oglWid, SLOT(scaleGridminus()));

	// Close mainwindow too, when OpenGL-viewer is closed
	connect(oglWid, SIGNAL(killAll()), this, SLOT(close()));

	// Solver has done his work
	connect(oglWid, SIGNAL(parserReady()), this, SLOT(parserMessageSlot()));

	// Some more commands from other widgets
	connect(oglWid, SIGNAL(updMaiWid()), this, SLOT(updMaiWid()));
	connect(oglWid, SIGNAL(updMaiInf()), this, SLOT(updMaiInf()));
	connect(aniEdi, SIGNAL(updMaiWid()), this, SLOT(updMaiWid()));
	connect(aniEdi, SIGNAL(updMaiInf()), this, SLOT(updMaiInf()));
	connect(aniEdi, SIGNAL(saveFromAni()), this, SLOT(saveAsSlot()));
	connect(fntEdi, SIGNAL(repaintMain()), this, SLOT(repaintSlot()));
	connect(fntEdi, SIGNAL(iconSize(const int)), this, SLOT(iconSizeSlot(const int)));

	// Bring on function editor finally
	functionSlot();
}


/* ********************************************************************** */
/* ACTIONS FOR MAINWINDOW-MENUS                                           */
/* ********************************************************************** */
void MainWindow::openSlot() {
	QString tmp=NAME;
	QString fileName=QFileDialog::getOpenFileName(
					this,
					"Open "+tmp+"-file",
					workDir,
					"Images (*.zhu)");

	// Got file or canceled file-selector?
	if(!fileName.isEmpty()) {

		// Check ending first
		if(!fileName.endsWith(".zhu"))
			fileName+=".zhu";

		// Load file then
		if(loadFile(fileName)) {
			aniEdi->animate(ani.active);
			emit gotNewFile();
		}
		else {
			QMessageBox::critical(this,
							tr("File opening\n"),
							tr("<b>Error</b> when loading zhu-file."
								"<p>Returning without action now!</p>")
			);
			return;
		}

		// Update current workDir
		fileName.resize(fileName.lastIndexOf('/'));
		workDir=fileName;

		// Show function editor anyhow
		funEdi->showNormal();
		funEdi->raise();
		funEdi->activateWindow();

		// Check animation-state at last
		if(ani.active)
			oglWid->setWindowTitle("Animation");
		else
			oglWid->setWindowTitle("Viewer");
	}
}

void MainWindow::saveAsSlot() {
	QString tmp=NAME;
	QString fileName=QFileDialog::getSaveFileName(
					this,
					"Save "+tmp+"-file",
					workDir,
					"Images (*.zhu)");

	// Got file or canceled file-selector?
	if(!fileName.isEmpty()) {

		// Check ending
		if(!fileName.endsWith(".zhu"))
			fileName+=".zhu";

		// Save and update current workDir
		if(!saveFile(fileName)) {
			QMessageBox::critical(this,
							tr("File saving\n"),
							tr("<b>Error</b> when saving zhu-file."
								"<p>Returning without action now!</p>")
			);
			return;
		}

		fileName.resize(fileName.lastIndexOf('/'));
		workDir=fileName;
	}
}

void MainWindow::exitSlot() {
	writeSettings();
	oglWid->close();
	ligEdi->close();
	matEdi->close();
	entEdi->close();
	aniEdi->close();
	picEdi->close();
	fntEdi->close();
	hlpEdi->close();
	solEdi->close();
	funEdi->close();
	#ifdef DEBUG
	errEdi->close();
	#endif
	close();
}

void MainWindow::helpSlot() {
	QString tmp=NAME;						// Just for automatic renaming
	QString tmpLow=tmp.toLower()+".html";
	char str[128];							// Just for automatic renaming
	QString content;

	QFile file(MYDOCDIR+tmpLow);
	QTextStream in(&file);

	tmpLow="Cannot open "+MYDOCDIR+tmpLow;
	q2cstrcpy(str, tmpLow);
	if(!file.open(QFile::ReadOnly | QFile::Text)) {
		QMessageBox::warning(this,
							tr(NAME),
							tr(str));
		return;
	}

	QApplication::setOverrideCursor(Qt::WaitCursor);
	content=in.readAll();
	QApplication::restoreOverrideCursor();

	hlpEdi->setWindowTitle(tmp+" help");	// Consistent name for help-window
	hlpEdi->setGeometry(30, 30, 660, 710);	// Give help a fixed position
	hlpEdi->setMinimumSize(300, 200);		// Give it a minimum size
	hlpEdi->insertHtml(content);
	hlpEdi->show();
}


/* ********************************************************************** */
/* SYSTEM INFO                                                            */
/* ********************************************************************** */
void MainWindow::systemSlot() {
	QString tmp=systemInfo();
	char *cstr;

	cstr=new char[tmp.length()+1];
	if(cstr!=NULL) {
		q2cstrcpy(cstr, tmp);
		QMessageBox::information(this, tr("System information"), tr(cstr));
		delete [] cstr;
		return;
	}

	delete [] cstr;
	QMessageBox::critical(this, tr("System information"), tr("Out of memory"));
}

void MainWindow::aboutSlot() {

	// Include compilation date
	const QString qstmp=(QString)(__DATE__);
	char tmp[20];

	q2cstrcpy(tmp, qstmp);
	QMessageBox::about(this,
		tr("About " NAME),
		tr(	"<b>" NAME "</b> is a an OpenGL-based 3D-function-viewer and solver."
			"<p>V " VERS "  compiled %1 </p>"	
			"<p>(c) 1991-2006 Heinz van Saanen</p>" ) .arg(tmp));
}

void MainWindow::lightSlot() {

	// Not yet activated?
	if(!ligEdi->isVisible())
		ligEdi->show();
	else
		ligEdi->showNormal();

	// Proceed with updates and bring widget to foreground
	ligEdi->updLigWid();
	ligEdi->raise();
	ligEdi->activateWindow();
}

void MainWindow::materialSlot() {

	// Not yet activated?
	if(!matEdi->isVisible())
		matEdi->show();
	else
		matEdi->showNormal();

	// Proceed with updates and bring widget to foreground
	matEdi->updMatWid();
	matEdi->raise();
	matEdi->activateWindow();
}

void MainWindow::animateSlot() {

	// Widget not yet activated?
	if(!aniEdi->isVisible())
		aniEdi->show();
	else
		aniEdi->showNormal();

	// Proceed with updates and bring widget to foreground
	aniEdi->updAniWid();
	aniEdi->raise();
	aniEdi->activateWindow();
}

void MainWindow::entireSlot() {

	// Not yet activated?
	if(!entEdi->isVisible())
		entEdi->show();
	else
		entEdi->showNormal();

	// Proceed with updates and bring widget to foreground
	entEdi->updEntWid();
	entEdi->raise();
	entEdi->activateWindow();
}

void MainWindow::solverSlot() {

	// Widget not yet activated?
	if(!solEdi->isVisible())
		solEdi->show();
	else
		solEdi->showNormal();

	// Bring widget to foreground
	solEdi->raise();
	solEdi->activateWindow();
}

void MainWindow::functionSlot() {

	// Widget not yet activated?
	if(!funEdi->isVisible())
		funEdi->show();
	else
		funEdi->showNormal();

	// Bring widget to foreground
	funEdi->updFunWid();
	funEdi->raise();
	funEdi->activateWindow();
}

void MainWindow::pictureSlot() {

	// Not yet activated?
	if(!picEdi->isVisible()) {
		picEdi->show();
		picEdi->updPicWid();
	}
	else
		picEdi->showNormal();

	// Bring widget to foreground
	picEdi->raise();
	picEdi->activateWindow();
}

void MainWindow::fontSlot() {

	// Not yet activated?
	if(!fntEdi->isVisible())
		fntEdi->show();
	else
		fntEdi->showNormal();

	// Bring widget to foreground
	fntEdi->raise();
	fntEdi->activateWindow();
}


// Bring all opened/minimized editors to front
void MainWindow::toFrontSlot() {

	// Lights
	if(ligEdi->isVisible()) {
		ligEdi->showNormal();
		ligEdi->raise();
		ligEdi->activateWindow();
	}

	// Materials
	if(matEdi->isVisible()) {
		matEdi->showNormal();
		matEdi->raise();
		matEdi->activateWindow();
	}

	// Global light
	if(entEdi->isVisible()) {
		entEdi->showNormal();
		entEdi->raise();
		entEdi->activateWindow();
	}

	// Animation
	if(aniEdi->isVisible()) {
		aniEdi->showNormal();
		aniEdi->raise();
		aniEdi->activateWindow();
	}

	// Solver report
	if(solEdi->isVisible()) {
		solEdi->showNormal();
		solEdi->raise();
		solEdi->activateWindow();
	}

	// Function editor
	if(funEdi->isVisible()) {
		funEdi->showNormal();
		funEdi->raise();
		funEdi->activateWindow();
	}
}


// Hide/minimize all opened editors
void MainWindow::toBackSlot() {

	// Lights
	if(ligEdi->isVisible())
		ligEdi->showMinimized();

	// Materials
	if(matEdi->isVisible())
		matEdi->showMinimized();

	// Global light
	if(entEdi->isVisible())
		entEdi->showMinimized();

	// Animation
	if(aniEdi->isVisible())
		aniEdi->showMinimized();

	// Solver report
	if(solEdi->isVisible())
		solEdi->showMinimized();

	// Function editor
	if(funEdi->isVisible())
		funEdi->showMinimized();
}


// Open all editors
void MainWindow::showAllSlot() {

	// Lights
	ligEdi->showNormal();
	ligEdi->raise();
	ligEdi->activateWindow();

	// Materials
	matEdi->showNormal();
	matEdi->raise();
	matEdi->activateWindow();

	// Global light
	entEdi->showNormal();
	entEdi->raise();
	entEdi->activateWindow();

	// Animation
	aniEdi->showNormal();
	aniEdi->raise();
	aniEdi->activateWindow();

	// Solver report
	solEdi->showNormal();
	solEdi->raise();
	solEdi->activateWindow();

	// Function editor
	funEdi->showNormal();
	funEdi->raise();
	funEdi->activateWindow();
}


// Initialize default viewing values
void MainWindow::resetViewSlot() {

	sta.angle=35.0f;
	sta.rotX=sta.rotY=sta.rotZ=0.0f;
	sta.transX=sta.transY=0.0f;
	sta.transZ=-5.0f;
	sta.crossX=sta.crossY=0.5f;

	updMaiWid();		// Update mainwindow widget
	updMaiInf();		// Update cross-section
	emit gotNewFile();	// Update viewer
}

// Initialize default scaling values
void MainWindow::resetScaleSlot() {

	sta.sa=sta.sx=sta.sy=sta.sz=1.0f;
	sta.spx=sta.spy=sta.spz=1.0f;

	updMaiWid();		// Update mainwindow widget
	updMaiInf();		// Update cross-section
	emit gotNewFile();	// Update viewer
}

// Initialize default scaling and viewing values
void MainWindow::resetAllSlot() {

	resetViewSlot();
	resetScaleSlot();

	updMaiWid();		// Update mainwindow widget
	updMaiInf();		// Update cross-section
}

// Toggle animation on/off
void MainWindow::runStopSlot() {
	ani.active = !ani.active;

	if(ani.active) {
		runStopAction->setIcon(QIcon(":/images/stop.png"));
		runStopAction->setToolTip("Stop animation");
		oglWid->setWindowTitle("Animation");
		aniEdi->animate(true);
	}
	else {
		runStopAction->setIcon(QIcon(":/images/play.png"));
		runStopAction->setToolTip("Start animation");
		oglWid->setWindowTitle("Viewer");
		aniEdi->animate(false);
	}
}


/* ********************************************************************** */
/* CREATE ACTIONS                                                         */
/* ********************************************************************** */
void MainWindow::createActions() {

	// FILE
	openAction = new QAction(QIcon(":/images/open.png"), tr("&Open..."), this);
	openAction->setShortcut(tr("Ctrl+O"));
	connect(openAction, SIGNAL(triggered()), this, SLOT(openSlot()));

	saveAsAction = new QAction(QIcon(":/images/saveas.png"), tr("&Save as"), this);
	saveAsAction->setShortcut(tr("Ctrl+S"));
	connect(saveAsAction, SIGNAL(triggered()), this, SLOT(saveAsSlot()));

	savePicAction = new QAction(QIcon(":/images/savepic.png"), tr("Save &pic"), this);
	savePicAction->setShortcut(tr("Ctrl+P"));
	connect(savePicAction, SIGNAL(triggered()), oglWid, SLOT(savePic()));

	exitAction = new QAction(QIcon(":/images/exit.png"), tr("E&xit"), this);
	exitAction->setShortcut(tr("Ctrl+X"));
	connect(exitAction, SIGNAL(triggered()), this, SLOT(exitSlot()));


	// EDITORS
	lightAction = new QAction(QIcon(":/images/light.png"), tr("Ligh&ts"), this);
	lightAction->setShortcut(tr("Ctrl+T"));
	connect(lightAction, SIGNAL(triggered()), this, SLOT(lightSlot()));

	materialAction = new QAction(QIcon(":/images/material.png"), tr("&Materials"), this);
	materialAction->setShortcut(tr("Ctrl+M"));
	connect(materialAction, SIGNAL(triggered()), this, SLOT(materialSlot()));

	animateAction = new QAction(QIcon(":/images/animate.png"), tr("&Animation"), this);
	animateAction->setShortcut(tr("Ctrl+A"));
	connect(animateAction, SIGNAL(triggered()), this, SLOT(animateSlot()));

	entireAction = new QAction(QIcon(":/images/entire.png"), tr("&Global light"), this);
	entireAction->setShortcut(tr("Ctrl+G"));
	connect(entireAction, SIGNAL(triggered()), this, SLOT(entireSlot()));

	solEdiAction = new QAction(QIcon(":/images/solve.png"), tr("Solver report"), this);
	connect(solEdiAction, SIGNAL(triggered()), this, SLOT(solverSlot()));

	funEdiAction = new QAction(QIcon(":/images/edit.png"), tr("Functions"), this);
	connect(funEdiAction, SIGNAL(triggered()), this, SLOT(functionSlot()));

	toFrontAction = new QAction(QIcon(":/images/tofront.png"), tr("Show all open"), this);
	connect(toFrontAction, SIGNAL(triggered()), this, SLOT(toFrontSlot()));

	toBackAction = new QAction(QIcon(":/images/toback.png"), tr("Hide all open"), this);
	connect(toBackAction, SIGNAL(triggered()), this, SLOT(toBackSlot()));

	showAllAction = new QAction(QIcon(":/images/viewall.png"), tr("Open all"), this);
	connect(showAllAction, SIGNAL(triggered()), this, SLOT(showAllSlot()));


	// SETTINGS
	pictureAction = new QAction(QIcon(":/images/savepic.png"), tr("Picture"), this);
	connect(pictureAction, SIGNAL(triggered()), this, SLOT(pictureSlot()));

	fineAction = new QAction(tr("fine"), this);
	fineAction->setCheckable(true);
	fineAction->setChecked(false);
	connect(fineAction, SIGNAL(triggered()), oglWid, SLOT(fineScale()));
	connect(fineAction, SIGNAL(triggered()), this,   SLOT(fineScale()));

	normalAction = new QAction(tr("normal"), this);
	normalAction->setCheckable(true);
	normalAction->setChecked(true);
	connect(normalAction, SIGNAL(triggered()), oglWid, SLOT(normalScale()));
	connect(normalAction, SIGNAL(triggered()), this,   SLOT(normalScale()));

	coarseAction = new QAction(tr("coarse"), this);
	coarseAction->setCheckable(true);
	coarseAction->setChecked(false);
	connect(coarseAction, SIGNAL(triggered()), oglWid, SLOT(coarseScale()));
	connect(coarseAction, SIGNAL(triggered()), this,   SLOT(coarseScale()));

	fontAction = new QAction(QIcon(":/images/fonts.png"), tr("Appearance"), this);
	connect(fontAction, SIGNAL(triggered()), this, SLOT(fontSlot()));


	// TOOLBARS
	zoomInAction = new QAction(QIcon(":/images/zoomin.png"), tr("Zoom in"), this);
	zoomInAction->setShortcut(tr("Ctrl++"));
	connect(zoomInAction, SIGNAL(triggered()), oglWid, SLOT(zoomIn()));

	zoomOutAction = new QAction(QIcon(":/images/zoomout.png"), tr("Zoom out"), this);
	zoomOutAction->setShortcut(tr("Ctrl+-"));
	connect(zoomOutAction, SIGNAL(triggered()), oglWid, SLOT(zoomOut()));

	transLeftAction = new QAction(QIcon(":/images/left.png"), tr("Move left"), this);
	transLeftAction->setShortcut(tr("Ctrl+L"));
	connect(transLeftAction, SIGNAL(triggered()), oglWid, SLOT(transLeft()));

	transRightAction = new QAction(QIcon(":/images/right.png"), tr("Move right"), this);
	transRightAction->setShortcut(tr("Ctrl+R"));
	connect(transRightAction, SIGNAL(triggered()), oglWid, SLOT(transRight()));

	transUpAction = new QAction(QIcon(":/images/up.png"), tr("Move up"), this);
	transUpAction->setShortcut(tr("Ctrl+U"));
	connect(transUpAction, SIGNAL(triggered()), oglWid, SLOT(transUp()));

	transDownAction = new QAction(QIcon(":/images/down.png"), tr("Move down"), this);
	transDownAction->setShortcut(tr("Ctrl+D"));
	connect(transDownAction, SIGNAL(triggered()), oglWid, SLOT(transDown()));

	resetViewAction = new QAction(QIcon(":/images/reset.png"), tr("View only"), this);
	connect(resetViewAction, SIGNAL(triggered()), this, SLOT(resetViewSlot()));

	resetScaleAction = new QAction(QIcon(":/images/reset.png"), tr("Scaling only"), this);
	connect(resetScaleAction, SIGNAL(triggered()), this, SLOT(resetScaleSlot()));

	resetAllAction = new QAction(QIcon(":/images/reset.png"), tr("All together"), this);
	connect(resetAllAction, SIGNAL(triggered()), this, SLOT(resetAllSlot()));

	solveAction = new QAction(QIcon(":/images/solve.png"), tr("Solve system"), this);
	connect(solveAction, SIGNAL(triggered()), oglWid, SLOT(solve()));

	runStopAction = new QAction(QIcon(":/images/play.png"), tr("Start/stop animation"), this);
	connect(runStopAction, SIGNAL(triggered()), this, SLOT(runStopSlot()));


	// HELP
	helpAction = new QAction(QIcon(":/images/help.png"), tr("Help"), this);
	helpAction->setShortcut(tr("F1"));
	connect(helpAction, SIGNAL(triggered()), this, SLOT(helpSlot()));

	systemAction = new QAction(QIcon(":/images/system.png"), tr("System info"), this);
	connect(systemAction, SIGNAL(triggered()), this, SLOT(systemSlot()));

	benchmarkAction = new QAction(QIcon(":/images/bench.png"), tr("&Benchmark"), this);
	benchmarkAction->setShortcut(tr("Ctrl+B"));
	connect(benchmarkAction, SIGNAL(triggered()), this, SLOT(benchmarkSlot()));

	aboutAction = new QAction(QIcon(":/images/icon.png"), tr("About Zhu3D"), this);
	connect(aboutAction, SIGNAL(triggered()), this, SLOT(aboutSlot()));

	aboutQtAction = new QAction(QIcon(":/images/qt-logo.png"), tr("About Qt"), this);
	connect(aboutQtAction, SIGNAL(triggered()), qApp, SLOT(aboutQt()));
}


/* ********************************************************************** */
/* CREATE/DELETE MENUS                                                    */
/* ********************************************************************** */
void MainWindow::createMenus() {

	// File
	fileMenu = menuBar()->addMenu(tr("&File"));
	fileMenu->addAction(openAction);
	fileMenu->addAction(saveAsAction);
	fileMenu->addSeparator();
	fileMenu->addAction(savePicAction);
	fileMenu->addSeparator();
	fileMenu->addAction(exitAction);

	// Editors
	editorMenu = menuBar()->addMenu(tr("&Editors"));
	editorMenu->addAction(lightAction);
	editorMenu->addAction(materialAction);
	editorMenu->addAction(entireAction);
	editorMenu->addSeparator();
	editorMenu->addAction(animateAction);
	editorMenu->addAction(solEdiAction);
	editorMenu->addAction(funEdiAction);
	editorMenu->addSeparator();
	editorMenu->addAction(toFrontAction);
	editorMenu->addAction(toBackAction);
	editorMenu->addSeparator();
	editorMenu->addAction(showAllAction);

	// Settings with submenus
	settingsMenu = menuBar()->addMenu(tr("Settings"));
	settingsMenu->addAction(pictureAction);
	settingsMenu->addSeparator();

	scaleMenu = settingsMenu->addMenu(QIcon(":/images/scale.png"), tr("Scaling"));
	scaleMenu->addAction(fineAction);
	scaleMenu->addAction(normalAction);
	scaleMenu->addAction(coarseAction);

	resetMenu = settingsMenu->addMenu(QIcon(":/images/reset.png"), tr("Reset"));
	resetMenu->addAction(resetViewAction);
	resetMenu->addAction(resetScaleAction);
	resetMenu->addAction(resetAllAction);

	settingsMenu->addSeparator();
	settingsMenu->addAction(fontAction);

	// System
	systemMenu = menuBar()->addMenu(tr("System"));
	systemMenu->addAction(systemAction);
	systemMenu->addAction(benchmarkAction);

	// Help
	helpMenu = menuBar()->addMenu(tr("&Help"));
	helpMenu->addAction(helpAction);
	helpMenu->addSeparator();
	helpMenu->addAction(aboutAction);
	helpMenu->addAction(aboutQtAction);
}


/* ********************************************************************** */
/* CREATE TOOLBARS                                                        */
/* ********************************************************************** */
void MainWindow::createToolBars() {

	// Get current iconsize from fntEdi
	QSize iconSize=fntEdi->getIcoSize();

	fileToolBar = addToolBar(tr("File"));
	fileToolBar->addAction(openAction);
	fileToolBar->addAction(saveAsAction);
	fileToolBar->setIconSize(iconSize);

	zoomToolBar = addToolBar(tr("Zoom"));
	zoomToolBar->addAction(zoomInAction);
	zoomToolBar->addAction(zoomOutAction);
	zoomToolBar->setIconSize(iconSize);

	transToolBar = addToolBar(tr("Move"));
	transToolBar->addAction(transLeftAction);
	transToolBar->addAction(transRightAction);
	transToolBar->addAction(transUpAction);
	transToolBar->addAction(transDownAction);
	transToolBar->setIconSize(iconSize);

	extraToolBar = addToolBar(tr("Extras"));
	extraToolBar->addAction(solveAction);
	extraToolBar->addAction(runStopAction);
	extraToolBar->setIconSize(iconSize);
}

// Like in menu-redrawing, again strange workaround is needed!!!
void MainWindow::iconSizeSlot(const int newSize) {
	QSize iconSize(QSize(newSize, newSize));

	// Destroy toolbars first
	delete fileToolBar;
	delete zoomToolBar;
	delete transToolBar;
	delete extraToolBar;

	// Recrate them
	createToolBars();

	// And now resize them!
	fileToolBar->setIconSize(iconSize);
	zoomToolBar->setIconSize(iconSize);
	transToolBar->setIconSize(iconSize);
	extraToolBar->setIconSize(iconSize);
}

/* ********************************************************************** */
/* MISCELLANEOUS GOODIES                                                  */
/* ********************************************************************** */
void MainWindow::readSettings() {
	QSettings settings("Heinz", NAME);
	QPoint pos=settings.value("MainWindowPos", QPoint(200, 200)).toPoint();
	QSize size=settings.value("MainWindowSize", QSize(400, 400)).toSize();
	QPoint solpos=settings.value("SolverEditPos", QPoint(200, 200)).toPoint();
	QSize solsize=settings.value("SolverEditSize", QSize(400, 400)).toSize();

	// Ensure to start with defaults on 1st program-start ever
	everStarted=false;
	everStarted=settings.value("FirstMainStart").toBool();
	if(everStarted==true) {
		workDir=settings.value("CurrentWorkDir").toString();
		resize(size);
		move(pos);
		solEdi->resize(solsize);
		solEdi->move(solpos);
	}
	else {
		everStarted=true;

		if(MHz==0.0f)
			QMessageBox::information(
				0,
				QObject::tr("Program start\n"),
				QObject::tr("<b>Hardware message:</b>"
							"<p>CPU-cycle-counter is not supported. "
							"Sys-info can not provide MhZ-values.</p>")
			);
		resize(QSize(420,470));
		move(QPoint(10, 10));
		solEdi->resize(QSize(700, 320));
		solEdi->move(QPoint(300, 350));
	}

}

void MainWindow::writeSettings() {
	QSettings settings("Heinz", NAME);

	// Save actual settings
	settings.setValue("MainWindowPos", pos());
	settings.setValue("MainWindowSize", size());
	settings.setValue("SolverEditPos", solEdi->pos());
	settings.setValue("SolverEditSize", solEdi->size());
	settings.setValue("CurrentWorkDir", workDir);
	settings.setValue("FirstMainStart", everStarted);

}

// Wire-mode+Points is not sensefull, so test if wired is checkable
void MainWindow::wireModeSlot() {

	// If mode == point
	if(sta.draw==2) {
		Wired->setChecked(false);
		Wired->setCheckable(false);
		return;
	}

	Wired->setCheckable(true);
	Wired->setChecked(sta.wired);
}

// Solver in oglWid has done his work
void MainWindow::parserMessageSlot() {
	QDate date=QDate::currentDate();
	QTime time=QTime::currentTime();

	solEdi->append("******************************************************");
	solEdi->append("Solver report: "+date.toString()+" at "+time.toString());
	solEdi->append("******************************************************");
	solEdi->append("F0(X,Y) = "+(QString)fun[0].str);
	solEdi->append("F1(X,Y) = "+(QString)fun[1].str);
	solEdi->append("F2(X,Y) = "+(QString)fun[2].str);
	solEdi->append("");
	solEdi->append(parseResult);
	solEdi->append("******************************************************");
	solEdi->append("\n");

	// Bring solver editor to foreground
	solverSlot();
}

// Show info; Oops, with ugly C-lib functions from old code:-)
void MainWindow::updMaiInf() {
	char message[100];
	char tmp[100];

	infoTextEdit->clear();

	// Current xy-value for actual cross-position
	if(fabs(sta.crossX*sta.sx)>1000000.0f || fabs(sta.crossX*sta.sx)<0.000001f)
		sprintf(message, "Cross X = %.5e ", sta.crossX*sta.sx);
	else
		sprintf(message, "Cross X = %.6f ", sta.crossX*sta.sx);
	infoTextEdit->append(message);

	if(fabs(sta.crossY*sta.sy)>1000000.0f || fabs(sta.crossY*sta.sy)<0.000001f)
		sprintf(message, "Cross Y = %.5e", sta.crossY*sta.sy);
	else
		sprintf(message, "Cross Y = %.6f", sta.crossY*sta.sy);
	infoTextEdit->append(message);


	// Calculate and show function values at cross-position
	strcpy(message, "F0(X,Y) = ");
	sta.crossZ=fxy0(sta.crossX, sta.crossY)/sta.sz;
	if(fabs(sta.crossZ*sta.sx)>1000000.0f || fabs(sta.crossZ*sta.sx)<0.000001f)
		sprintf(tmp, "%.5e", sta.crossZ);
	else
		sprintf(tmp, "%.6f", sta.crossZ);
	strcat(message, tmp);
	infoTextEdit->append(message);

	strcpy(message, "F1(X,Y) = ");
	sta.crossZ=fxy1(sta.crossX, sta.crossY)/sta.sz;
	if(fabs(sta.crossZ*sta.sx)>1000000.0f || fabs(sta.crossZ*sta.sx)<0.000001f)
		sprintf(tmp, "%.5e", sta.crossZ);
	else
		sprintf(tmp, "%.6f", sta.crossZ);
	strcat(message, tmp);
	infoTextEdit->append(message);

	strcpy(message, "F2(X,Y) = ");
	sta.crossZ=fxy2(sta.crossX, sta.crossY)/sta.sz;
	if(fabs(sta.crossZ*sta.sx)>1000000.0f || fabs(sta.crossZ*sta.sx)<0.000001f)
		sprintf(tmp, "%.5e", sta.crossZ);
	else
		sprintf(tmp, "%.6f", sta.crossZ);
	strcat(message, tmp);
	infoTextEdit->append(message);
}


/* ********************************************************************** */
/* FILE STUFF                                                             */
/* ********************************************************************** */
// Load
bool MainWindow::loadFile(const QString &fileName) {
	int i, j, k;
	QFile file(fileName);
	QString tmp;

	// Check for startup-file only
	if(!file.exists())
		return false;

	// Open save
	if(!file.open(QFile::ReadOnly | QFile::Text)) {
		QMessageBox::warning(this,
							tr(NAME),
							tr("Cannot read file %1:\n%2.")
							.arg(fileName)
							.arg(file.errorString()));
		return false;
	}

	QTextStream in(&file);
	QApplication::setOverrideCursor(Qt::WaitCursor);

	// Functions & parametric system
	tmp=file.readLine(); tmp=tmp.remove(QChar('\n')); q2cstrcpy(fun[0].str, tmp);
	tmp=file.readLine(); tmp=tmp.remove(QChar('\n')); q2cstrcpy(fun[1].str, tmp);
	tmp=file.readLine(); tmp=tmp.remove(QChar('\n')); q2cstrcpy(fun[2].str, tmp);
	tmp=file.readLine(); tmp=tmp.remove(QChar('\n')); q2cstrcpy(par.str[0], tmp);
	tmp=file.readLine(); tmp=tmp.remove(QChar('\n')); q2cstrcpy(par.str[1], tmp);
	tmp=file.readLine(); tmp=tmp.remove(QChar('\n')); q2cstrcpy(par.str[2], tmp);

	in >> fun[0].drawFun >> fun[1].drawFun >> fun[2].drawFun;
	in >> par.drawPar;
	in >> par.sLow >> par.sHig >> par.tLow >> par.tHig;
	in >> sta.spx >> sta.spy >> sta.spz;
	in >> sta.parMod;

	// Status rest
	in >> sta.angle;
	in >> sta.rotX >> sta.rotY >> sta.rotZ;
	in >> sta.transX >> sta.transY >> sta.transZ;
	in >> sta.crossX >> sta.crossY;
	in >> sta.sa >> sta.sx >> sta.sy >> sta.sz;
	in >> sta.grids;
	in >> sta.tessWidth;

	in >> sta.cross >> sta.axes >> sta.wired;
	in >> sta.draw;
	in >> sta.ediLig >> sta.ediMat;

	// Lights
	in >> sta.ediLig;
	for(i=0; i<MAXLIG; i++) {
		in >> lig[i].setLig;
		in >> lig[i].ambLigLoc;
		in >> lig[i].difLigLoc;
		in >> lig[i].speLigLoc;
		in >> lig[i].ligDis;
		for(j=0; j<3; j++) in >> lig[i].ligPos[j];
		for(j=0; j<4; j++) in >> lig[i].ambLig[j];
		for(j=0; j<4; j++) in >> lig[i].difLig[j];
		for(j=0; j<4; j++) in >> lig[i].speLig[j];
		for(j=0; j<4; j++) in >> lig[i].spoDir[j];
		in >> lig[i].spoSta;
		in >> lig[i].spoExp;
		in >> lig[i].spoCut;
		in >> lig[i].conAtt;
		in >> lig[i].linAtt;
		in >> lig[i].quaAtt;
	}

	// Materials
	in >> sta.ediMat;
	for(k=0; k<2; k++)
	for(i=0; i<MAXFUN+1; i++) {
		in >> mat[i].ambMatLoc[k];
		in >> mat[i].difMatLoc[k];
		in >> mat[i].speMatLoc[k];
		in >> mat[i].emiMatLoc[k];
		for(j=0; j<4; j++) in >> mat[i].ambMat[k][j];
		for(j=0; j<4; j++) in >> mat[i].difMat[k][j];
		for(j=0; j<4; j++) in >> mat[i].speMat[k][j];
		for(j=0; j<4; j++) in >> mat[i].emiMat[k][j];
		in >> mat[i].shiMat[k];
	}
	for(i=0; i<MAXFUN+1; i++) in >> mat[i].twoSid;

	// Entire
	in >> ent.ambEntLoc;
	in >> ent.bacEntLoc;
	in >> ent.modInf;
	in >> ent.modOne;
	for(i=0; i<4; i++) in >> ent.bacLig[i];
	for(i=0; i<4; i++) in >> ent.modLig[i];

	// Animation
	in >> ani.xFac >> ani.yFac >> ani.zFac;
	in >> ani.locXYZ;
	in >> ani.active;
	in >> ani.fps;
	if(ani.fps==0)		// Consider zuh3-files <= 1.4
		ani.fps=50;		// Avoid zero division in animation-part

	// Perform updates
	updMaiWid();
	updMaiInf();
	ligEdi->updLigWid();
	matEdi->updMatWid();
	entEdi->updEntWid();
	aniEdi->updAniWid();
	funEdi->updFunWid();

	// Clean up
	QApplication::restoreOverrideCursor();
	setCurrentFile(fileName);
	file.close();
	return true;
}

// Save
bool MainWindow::saveFile(const QString &fileName) {
	int i, j, k;
	QFile file(fileName);

	// Open save
	if(!file.open(QFile::WriteOnly | QFile::Text)) {
        QMessageBox::warning(this,
							tr(NAME),
							tr("Cannot write file %1:\n%2.")
							.arg(fileName)
							.arg(file.errorString()));
		return false;
	}

	QTextStream out(&file);
	QApplication::setOverrideCursor(Qt::WaitCursor);

	// Functions & parametric system
	out << fun[0].str << "\n" << fun[1].str << "\n" << fun[2].str << "\n";
	out << par.str[0] << "\n" << par.str[1] << "\n" << par.str[2] << "\n";

	out << fun[0].drawFun << "\n" << fun[1].drawFun << "\n" << fun[2].drawFun << "\n";
	out << par.drawPar << "\n";
	out << par.sLow << "\n" << par.sHig << "\n" << par.tLow << "\n" << par.tHig << "\n";
	out << sta.spx << "\n" << sta.spy << "\n" << sta.spz << "\n";
	out << sta.parMod << "\n";

	// Status rest
	out << sta.angle << "\n";
	out << sta.rotX << "\n" << sta.rotY << "\n" << sta.rotZ << "\n";
	out << sta.transX << "\n" << sta.transY << "\n" << sta.transZ << "\n";
	out << sta.crossX << "\n" << sta.crossY << "\n";
	out << sta.sa << "\n" << sta.sx << "\n" << sta.sy << "\n" << sta.sz << "\n";
	out << sta.grids << "\n";
	out << sta.tessWidth << "\n";

	out << sta.cross << "\n" << sta.axes << "\n" << sta.wired << "\n";
	out << sta.draw << "\n";
	out << sta.ediLig << "\n" << sta.ediMat << "\n";

	// Lights
	out << sta.ediLig << "\n";
	for(i=0; i<MAXLIG; i++) {
		out << lig[i].setLig << "\n";
		out << lig[i].ambLigLoc << "\n";
		out << lig[i].difLigLoc << "\n";
		out << lig[i].speLigLoc << "\n";
		out << lig[i].ligDis << "\n";
		for(j=0; j<3; j++) out << lig[i].ligPos[j] << "\n";
		for(j=0; j<4; j++) out << lig[i].ambLig[j] << "\n";
		for(j=0; j<4; j++) out << lig[i].difLig[j] << "\n";
		for(j=0; j<4; j++) out << lig[i].speLig[j] << "\n";
		for(j=0; j<4; j++) out << lig[i].spoDir[j] << "\n";
		out << lig[i].spoSta << "\n";
		out << lig[i].spoExp << "\n";
		out << lig[i].spoCut << "\n";
		out << lig[i].conAtt << "\n";
		out << lig[i].linAtt << "\n";
		out << lig[i].quaAtt << "\n";
	}

	// Materials
	out << sta.ediMat << "\n";
	for(k=0; k<2; k++)
	for(i=0; i<MAXFUN+1; i++) {
		out << mat[i].ambMatLoc[k] << "\n";
		out << mat[i].difMatLoc[k] << "\n";
		out << mat[i].speMatLoc[k] << "\n";
		out << mat[i].emiMatLoc[k] << "\n";
		for(j=0; j<4; j++) out << mat[i].ambMat[k][j] << "\n";
		for(j=0; j<4; j++) out << mat[i].difMat[k][j] << "\n";
		for(j=0; j<4; j++) out << mat[i].speMat[k][j] << "\n";
		for(j=0; j<4; j++) out << mat[i].emiMat[k][j] << "\n";
		out << mat[i].shiMat[k] << "\n";
	}
	for(i=0; i<MAXFUN+1; i++)
		out << mat[i].twoSid << "\n";

	// Entire
	out << ent.ambEntLoc << "\n";
	out << ent.bacEntLoc << "\n";
	out << ent.modInf << "\n";
	out << ent.modOne << "\n";
	for(i=0; i<4; i++) out << ent.bacLig[i] << "\n";
	for(i=0; i<4; i++) out << ent.modLig[i] << "\n";

	// Animation
	out << ani.xFac << "\n" << ani.yFac << "\n" << ani.zFac << "\n";
	out << ani.locXYZ << "\n";
	out << ani.active << "\n";
	out << ani.fps << "\n";

	// Clean up
	QApplication::restoreOverrideCursor();
	setCurrentFile(fileName);
	file.close();
	return true;
}

// Prepare filename as current window-name
void MainWindow::setCurrentFile(const QString &fileName) {
	currentFile = fileName;

	QString shownName;
	if (currentFile.isEmpty())
		shownName = "untitled.txt";
	else
		shownName = strippedName(currentFile);

	setWindowTitle(tr("%1[*] - %2").arg(shownName).arg(tr(NAME)));
}

QString MainWindow::strippedName(const QString &fullFileName) {
	return QFileInfo(fullFileName).fileName();
}


/* ********************************************************************** */
/* HANDLE CHECKBOXES FOR SUBMENUS FILEFORMAT AND SCALING                  */
/* ********************************************************************** */
void MainWindow::fineScale() {
	if(fineChecked) {
		fineAction->setChecked(true);
		return;
	}
	fineChecked=true;
	fineAction->setChecked(true);
	normalChecked=false;
	normalAction->setChecked(false);
	coarseChecked=false;
	coarseAction->setChecked(false);
}

void MainWindow::normalScale() {
	if(normalChecked) {
		normalAction->setChecked(true);
		return;
	}
	fineChecked=false;
	fineAction->setChecked(false);
	normalChecked=true;
	normalAction->setChecked(true);
	coarseChecked=false;
	coarseAction->setChecked(false);
}

void MainWindow::coarseScale() {
	if(coarseChecked) {
		coarseAction->setChecked(true);
		return;
	}
	fineChecked=false;
	fineAction->setChecked(false);
	normalChecked=false;
	normalAction->setChecked(false);
	coarseChecked=true;
	coarseAction->setChecked(true);
}


/* ********************************************************************** */
/* HANDLE MAIN-WIDGET REDRAW WHEN NEW FILE IS LOADED                      */
/* ********************************************************************** */
void MainWindow::updMaiWid() {

	// Show
	Function0->setChecked(fun[0].drawFun);
	Ui_mainCtrl::Function1->setChecked(fun[1].drawFun);
	Ui_mainCtrl::Function2->setChecked(fun[2].drawFun);
	Ui_mainCtrl::Function3->setChecked(par.drawPar);
	Ui_mainCtrl::Cross->setChecked(sta.cross);
	Ui_mainCtrl::Axes->setChecked(sta.axes);

	// Mode
	Ui_mainCtrl::Point->setChecked(sta.draw==2?1:0);
	Ui_mainCtrl::Triangle->setChecked(sta.draw==0?1:0);
	Ui_mainCtrl::Quad->setChecked(sta.draw==1?1:0);
	Ui_mainCtrl::Parameter->setChecked(sta.parMod==1?1:0);
	wireModeSlot();

	// Rotation-sliders
	Ui_mainCtrl::xSlider->setValue((int)sta.rotX);
	Ui_mainCtrl::ySlider->setValue((int)sta.rotY);
	Ui_mainCtrl::zSlider->setValue((int)sta.rotZ);

	// We are in function-mode
	if(!sta.parMod) {
		solveAction->setEnabled(true);
		xscaleLabel->setText(" X");
		xscaleLabel->setToolTip("X-values");
		yscaleLabel->setText(" Y");
		yscaleLabel->setToolTip("Y-values");
		zscaleLabel->setText(" Z");
		zscaleLabel->setToolTip("Z-values");
		xyzPlus->setEnabled(true);
		xyzMinus->setEnabled(true);
		xyzscaleLabel->setEnabled(true);
		xyzscaleLabel->setToolTip("All X+Y+Z together");
		infoTextEdit->setEnabled(true);
		funEdi->setWindowTitle("Function editor");
		funEdiAction->setText("Functions");
		Parameter->setToolTip("Enter parameter-mode and leave function-mode");
	}
	// Or in paramter mode
	else {
		solveAction->setEnabled(false);
		xscaleLabel->setText(" U");
		xscaleLabel->setToolTip("U-values");
		yscaleLabel->setText(" V");
		yscaleLabel->setToolTip("V-values");
		zscaleLabel->setText(" A");
		zscaleLabel->setToolTip("All U+V together");
		xyzPlus->setEnabled(false);
		xyzMinus->setEnabled(false);
		xyzscaleLabel->setEnabled(false);
		infoTextEdit->setEnabled(false);
		funEdi->setWindowTitle("Parameter editor");
		funEdiAction->setText("Parameters");
		Parameter->setToolTip("Leave parameter-mode and enter function-mode");
	}

	//Toggle labels on animation
	if(!ani.active) {
		if(!sta.parMod) {
			solveAction->setEnabled(true);
			infoTextEdit->setEnabled(true);
		}
		benchmarkAction->setEnabled(true);
		runStopAction->setIcon(QIcon(":/images/play.png"));
		runStopAction->setToolTip("Start animation");
	}
	else {
		solveAction->setEnabled(false);
		infoTextEdit->setEnabled(false);
		benchmarkAction->setEnabled(false);
		runStopAction->setIcon(QIcon(":/images/stop.png"));
		runStopAction->setToolTip("Stop animation");
	}
}


/* ********************************************************************** */
/* BENCHMARK                                                              */
/* ********************************************************************** */
void MainWindow::benchmarkSlot() {
	const GLfloat roffset=0.25f;
	const int mSec=5000;
	int frames=0;

	char message[50];
	QString result("Benchmark ran "), tmp;
	QString currentTitle=this->windowTitle();

	// Backup and load benchmark-file
	saveToRam();
	if(!loadFile(MYSYSDIR+"benchmark.zhu")) {
		QMessageBox::critical(this, tr("Benchmark"), tr("Can't open benchmark.zhu"));
		restoreFromRam();
		return;
	}

	// Resize OpenGL-widget
	oglWid->writeSettings();
	oglWid->move(QPoint(0,0));
	oglWid->resize(QSize(700,700));
	oglWid->setWindowTitle("Benchmarking now ...");
	oglWid->raise();
	oglWid->activateWindow();
	emit gotNewFile();

	// Perform demo/benchmark
	benchReady=false;
	QApplication::setOverrideCursor(Qt::WaitCursor);
	QTimer::singleShot(mSec, this, SLOT(benchTimerSlot()));
	while(!benchReady) {
			sta.rotX+=roffset;
			sta.rotY+=3*roffset;
			sta.rotZ+=4*roffset;
			emit gotNewView();
			frames++;
			QApplication::processEvents();		// Give timer a chance!
	}
	QApplication::restoreOverrideCursor();

	// Prepare message-string
	tmp.setNum(mSec/1000.0, 'f', 0);
	result+=tmp+" sec.\nFPS = ";
	tmp.setNum(frames/(mSec/1000.0), 'f', 2);
	result+=tmp;

	// Show benchmark-info
	q2cstrcpy(message, result);
	QMessageBox::information(this,
		tr("Benchmark"),
		tr(message)
	);

	// Restore and perform updates
	restoreFromRam();
	oglWid->setWindowTitle("Viewer");
	oglWid->readSettings();
	this->setWindowTitle(currentTitle);
	this->raise();
	this->activateWindow();
	updMaiWid();
	updMaiInf();
	ligEdi->updLigWid();
	matEdi->updMatWid();
	entEdi->updEntWid();
	aniEdi->updAniWid();
	emit gotNewFile();
}
