/****************************************************************************
**
** Copyright (C) 1991-2006 Heinz van Saanen
**
** This file is part of the the function viewer Zhu3D.
**
** This file may be used under the terms of the GNU General Public
** License version 2.0 as published by the Free Software Foundation
** and appearing in the file LICENSE.GPL included in the packaging of
** this file.
**
** This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
** WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
**
****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <math.h>
#include "parser.h"


// Uncomment for NPV-calculations in 2D-mode
//#define TWODIM


// Globals
// ============================================================================
FTYPE x,y;			// Global variables for functions
FTYPE s,t;			// Global variables for parametric-mode
int elevel;			// If elevel<>0 an error occured
#ifdef TWODIM
int npvloaded;		// Flag for loading NPV()-File successfully
#endif

char *errorStr[] = {
	"ERROR 1: Division by Zero",
	"ERROR 2: Modulus is Zero",
	"ERROR 3: Unknown Expression",
	"ERROR 4: Faculty is negative",
	"ERROR 5: Root is negative",
	"ERROR 6: Logarithm is negative",
	"ERROR 7: Brackets don't match",
	"ERROR 8: Faculty is too big"
};


// Function & variable definintions
// ============================================================================
static char *cp;					// Global char-pointer for parsed String
static FTYPE expression(void);
static FTYPE term(void);
static void factor(FTYPE *value);
static FTYPE fak(FTYPE value);
#ifdef TWODIM
static FTYPE NPV(FTYPE kzf);
#endif


// Central parsing-function
// ============================================================================
FTYPE parse(const char *string) {
	cp=(char *)string;				// Init global "parse" pointer
	elevel=0;						// Error-level is zero initially
	if(*cp=='#') return (0.0f);	// Quit calculations on comment
	return (expression());
}


// Calculating expression
// ============================================================================
static FTYPE expression(void) {
	FTYPE value;

	value=term();
	if(elevel) return 0.0f;
	while((*cp=='+') || (*cp=='-'))
		if(*cp++=='+') {
			value+=term();
			if(elevel) return 0.0f;
		}
		else {
			value-=term();
			if(elevel) return 0.0f;
		}
	return (value);
}


// Calculating terms
// ============================================================================
static FTYPE term(void) {
	FTYPE value, help;

	factor(&value);
	if(elevel) return 0.0f;
	while(*cp=='*' || *cp=='/' || isalpha(*cp) || *cp=='(')
		if(*cp=='*') {
			cp++;
			factor(&help);
			if(elevel) return (0.0f);
			value*=help;
		}
		else if(isalpha(*cp) || *cp=='(') {
			factor(&help);
			if(elevel) return (0.0f);
			value*=help;
		}
		else {
			cp++;
			factor(&help);
			if(elevel) return (0.0f);
			if(help==0.0f) { elevel=1; show_error(elevel); return 0.0f; }
			value/=help;
		}
	return (value);
}


// Handling unitary operators like ! or ^, constants, x2, sin(x)2, ...
// ============================================================================
#define CHECK(); \
if(*cp=='!') { *value=fak(*value); cp++; } \
if(*cp=='^') { cp++; factor(&help); *value=pow(*value,help); } \
if(*cp=='%') { cp++; factor(&help); if(help>0.0f) *value=fmod(*value,help); \
else { elevel=2; }} \
if(isdigit(*cp) || (*cp=='$')) *value*=expression();


// Calculating Factor-Values
// ============================================================================
static void factor(FTYPE *value) {
	char doublestr[30];				// Memory for number-extraction
	char *dspt=doublestr; 			// Pointer to actual position in doublestring
	FTYPE help;

	// Numbers
	if(elevel) { *value=0.0f; return; }
	if(isdigit(*cp) || *cp=='.') {					// Starting a number?
		while(isdigit(*cp)||*cp=='.')				// Then copy until not number or decimalpoint
			*dspt++=*cp++;
		if(*cp=='E' && !isalpha(*(cp+1))) {			// Exponent following?
			*dspt++=*cp++;							// Then go on copying
			if(*cp=='+'||*cp=='-'||isdigit(*cp)) {	// Check Exponent-Sign
	       	*dspt++=*cp++;							// And go on Copying
				while(isdigit(*cp)) *dspt++=*cp++;	// Until Exponent-Numbers finished
			}
		}
		*dspt='\0';								// Close "number-part" of string
		*value=atof(doublestr);					// Float conversion finished
		CHECK();								// Check for following unitary Operator ! , ^ and %
	}											// From here on check everthing not a Number

	// Basic operators
	else	if(*cp=='-') {						// Handles leading Minus
				cp++;
				factor(&help);
				*value=-help;
			}
	else	if(*cp=='+') {						// Handles leading Plus
				cp++;
				*value=expression();
				cp++;
			}
	else	if(*cp=='(') {						// Handles open Bracket
				cp++;
				*value=expression();
				cp++;
				CHECK();						// Check following unitary Operator ! , ^ and %
			}
	else	if(!strncmp("X",cp,1)) {			// Handles global X
				cp++;
				*value=x;
				CHECK();
			}
	else	if(!strncmp("Y",cp,1)) {			// Handles global Y
				cp++;
				*value=y;
				CHECK();
			}

	// Trigonometric functions
	else	if(!strncmp("SIN(",cp,4)) {			// Handles sin(...)
				cp+=4;
				*value=sin(expression());
				cp++;
				CHECK();
			}
	else	if(!strncmp("COS(",cp,4)) {			// Handles cos(...)
				cp+=4;
				*value=cos(expression());
				cp++;
			}
	else	if(!strncmp("TAN(",cp,4)) {			// Handles tan(...)
				cp+=4;
				*value=tan(expression());
				cp++;
			}
	else	if(!strncmp("ASIN(",cp,5)) {		// Handles asin(...)
				cp+=5;
				*value=asin(expression());
				cp++;
				CHECK();
			}
	else	if(!strncmp("ACOS(",cp,5)) {		// Handles acos(...)
				cp+=5;
				*value=acos(expression());
				cp++;
				CHECK();
			}
	else	if(!strncmp("ATAN(",cp,5)) {		// Handles atan(...)
				cp+=5;
				*value=atan(expression());
				cp++;
				CHECK();
			}

	// Hyperbolic functions
	else	if(!strncmp("SINH(",cp,5)) {		// Handles sinh(...)
				cp+=5;
				*value=sinh(expression());
				cp++;
				CHECK();
			}
	else	if(!strncmp("COSH(",cp,5)) {		// Handles cosh(...)
				cp+=5;
				*value=cosh(expression());
				cp++;
				CHECK();
			}
	else	if(!strncmp("TANH(",cp,5)) {		// Handles tanh(...)
				cp+=5;
				*value=tanh(expression());
				cp++;
				CHECK();
			}
	else	if(!strncmp("ASINH(",cp,6)) {		// Handles asinh(...)
				cp+=6;
				*value=asinh(expression());
				cp++;
				CHECK();
			}
	else	if(!strncmp("ACOSH(",cp,6)) {		// Handles acosh(...)
				cp+=6;
				*value=acosh(expression());
				cp++;
				CHECK();
			}
	else	if(!strncmp("ATANH(",cp,6)) {		// Handles atanh(...)
				cp+=6;
				*value=atanh(expression());
				cp++;
				CHECK();
			}

	// Miscellaneous functions
	else	if(!strncmp("SQRT(",cp,5)) {		// Handles sqrt(...)
				cp+=5;
				*value=expression();
				if(*value>=0.0f) *value=sqrt(*value); else { elevel=5; *value=0.0f; }
				cp++;
				CHECK();
			}
	else	if(!strncmp("ABS(",cp,4)) {			// Handles abs(...)
				cp+=4;
				*value=fabs(expression());
				cp++;
				CHECK();
			}
	else	if(!strncmp("SIG(",cp,4)) {			// Handles signum(...)
				cp+=4;
				*value=(expression()>=0.0f)?1.0f:-1.0f;
				cp++;
				CHECK();
			}
	else	if(!strncmp("INT(",cp,4)) {			// Handles int(...)
				cp+=4;
				*value=(int)(expression());
				cp++;
				CHECK();
			}
	else	if(!strncmp("FRAC(",cp,5)) {		// Handles frac(...)
				cp+=5;
				*value=expression();
				*value-=(int)(*value);
				cp++;
				CHECK();
			}
	else	if(!strncmp("LOG(",cp,4)) {			// Handles natural log(...)
				cp+=4;
				*value=expression();
				if(*value>0.0f) *value=log(*value); else { elevel=6; *value=0.0f; }
				cp++;
			}
	else	if(!strncmp("LOG10(",cp,6)) {		// Handles log10(...)
				cp+=6;
				*value=expression();
				if(*value>0.0f) *value=log10(*value); else { elevel=6; *value=0.0f; }
				cp++;
			}
	#ifdef TWODIM
	else	if(!strncmp("NPV(",cp,4)) {			// Handles Net Present Value (NPV)
				cp+=4;							// For Solving Equations of
				if(npvloaded==true)			// Internal Rates of Return (IRR)
				   *value=NPV(expression());	// Used in 2D-Version of Zhu3D only
				else
				   *value=0.0f;
				cp++;
			}
	#endif
	else	if(!strncmp("GAMMA(",cp,6)) {		// Handles gamma(...)
				cp+=6;
				*value=expression();
				*value=tgamma(*value);
				cp++;
			}

	// Constants
	else	if(!strncmp("$PI",cp,3)) {			// Handles PI=3.1415...
				cp+=3;							// Offset just 2 because of constant Value
				*value=MY_PI;					// Constant! don't increment Pointer
				CHECK();
			}
	else	if(!strncmp("$E",cp,2)) {			// Handles Euler-Constant E=2.7182...
				cp+=2;							// Offset just 5 because of constant Value
				*value=MY_E;					// Constant! don't increment Pointer
				CHECK();
			}

	// Parametric calculations wtih s and t
	else	if(!strncmp("S",cp,1)) {			// Handles global S
				cp++;
				*value=s;
				CHECK();
			}
	else	if(!strncmp("T",cp,1)) {			// Handles global T
				cp++;
				*value=t;
				CHECK();
			}

	// Error
	else	{									// Else something "strange" occured
		elevel=3;
		show_error(elevel);
		return;
	}
} // End of factor()


/* ************************************************************************* */
/*  Some utility-functions                                                   */
/* ************************************************************************* */

// Calculates faculty. Returns 1.0 for parameter==0
// ============================================================================
static FTYPE fak(FTYPE value) {
	long long i, j;
	FTYPE k=1.0;
	j=(long long)(value);

	if(j<0) {
		elevel=4;
		return 0.0f;
	}

	if(j>20) {
		elevel=8;
		return 0.0f;
	}

	for(i=2; i<=j; i++)
		k*=(FTYPE)i;
	
	return (k);
}


// For some Cashflow-Analysis; used in 2D-Version of Zhu3D only
// ===========================================================
#ifdef TWODIM
#define INDEX 256			// Maximum Index for cashflows to read
FTYPE inpt[INDEX][2];		// Field for discounted Cash-Flows
int cfn;					// Number of Cash-Flows for NPV()
int line;					// "Pointer" to actual Line of File-Input
int maxline=0;				// Lines for Input-File


// Calculates with Data in inpt [0..cfn-1][1]
// kzf = Interest Rate for discounted Cashflows
static FTYPE NPV(FTYPE kzf) {
FTYPE su=inpt[0][1], help1;
int j;
	if((kzf<=-100.0)) return 0.0f;
	kzf/=100.0;
	for(j=1; j<cfn; j++) {
		help1=pow(1.0+kzf,j);
		if(help1==0) return 0.0f;
		help1=1.0/help1;
		su+=inpt[j][1]*help1;
	}
	return su;
}


// Loading NPV-File; used in 2D-Version of Zhu3D only
// ===========================================================
void opencsv(char *s) {
FILE *fp;
int STRLEN=80;
char c1[STRLEN];

	npvloaded=false;
	for(line=0; line<INDEX; line++) inpt[line][1]='\0';
	line=0;

	if((fp=fopen(s,"r"))==NULL) {
		printf("Unable to open file %s\n", s);
		return;
	}
	else {
		while(fgets(&c1[line], STRLEN, fp) != NULL) {
			inpt[line][1]=atof(&c1[line]);
			line++;
		}
		npvloaded=true;
		printf("File %s loaded\n", s);
	}
	cfn=maxline=line;
	line=0;
	fclose(fp);
}
// TWODIM
#endif

// Error-Handling
// ===========================================================
void show_error(const int number) {
	switch (number) {
		case 1: printf("ERROR 1: Division by Zero\n"); break;
		case 2: printf("ERROR 2: Modulus is Zero\n"); break;
		case 3: printf("ERROR 3: Unknown Expression\n"); break;
		case 4: printf("ERROR 4: Faculty is negative\n"); break;
		case 5: printf("ERROR 5: Root is negative\n"); break;
		case 6: printf("ERROR 6: Logarithm is negative\n"); break;
		case 7: printf("ERROR 7: Brackets don't match\n"); break;
		case 8: printf("ERROR 8: Faculty is too big\n"); break;
		default: break;
	}
}
