/****************************************************************************
**
** Copyright (C) 1991-2006 Heinz van Saanen
**
** This file is part of the the function viewer Zhu3D.
**
** This file may be used under the terms of the GNU General Public
** License version 2.0 as published by the Free Software Foundation
** and appearing in the file LICENSE.GPL included in the packaging of
** this file.
**
** This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
** WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
**
****************************************************************************/

#include "property.h"


// Extern globals
LIG lig[MAXLIG];	// Light properties
MAT mat[MAXFUN+1];	// Material properties
ENT ent;			// Entire OpenGL-properties
ANI ani;			// Animation properties
FUN fun[MAXFUN];	// The functions
PAR par;			// Parameter properties
STA sta;			// General status-values
PIC pic;			// Picture-dimensions


// Constant values
const CON con={		// Constant factors
		1.500f,		// Zoom-factor
		0.050f,		// Translation-factor
		1.003f,		// Fine scale-factor
		1.050f,		// Normal scale-factor
		1.200f,		// Coarse scale-factor
		1.250f		// Tesselation-factor
};


// For backup to RAM
static LIG blig[MAXLIG];		// Light properties
static MAT bmat[MAXFUN+1];		// Material properties
static ENT bent;				// Entire OpenGL-properties
static ANI bani;				// Animation properties
static FUN bfun[MAXFUN];		// The functions
static PAR bpar;				// Parameter properties
static STA bsta;				// General status-values

void saveToRam();
void restoreFromRam();


/* *************************************************** */
/* Copy a QString to a plain old character-string      */
/* *************************************************** */
void q2cstrcpy(char *str, const QString qstr) {
	int i;

	QByteArray qba=qstr.toAscii();
	for(i=0; i<qba.size(); i++)
		str[i]=qba.at(i);
	str[i]='\0';
}


/* *************************************************** */
/* INITIALIZE LIGHTS WITH DEFAULT-VALUES               */
/* *************************************************** */
void iniLig() {
	int i, j;

	// Set common values
	for(i=0; i<MAXLIG; i++) {
		// Set all alpha-channels 1.0f
		lig[i].ambLig[3]=1.0f;
		lig[i].difLig[3]=1.0f;
		lig[i].speLig[3]=1.0f;

		// Booleans
		lig[i].setLig=false;
		lig[i].ambLigLoc=true;
		lig[i].difLigLoc=true;
		lig[i].speLigLoc=true;

		// Colour settings
		for(j=0; j<3; j++) lig[i].ambLig[j]=0.1f;
		for(j=0; j<3; j++) lig[i].difLig[j]=1.0f;
		for(j=0; j<3; j++) lig[i].speLig[j]=0.25f;
		for(j=0; j<3; j++) lig[i].spoDir[j]=0.0f;

		// Spotlight
		lig[i].spoSta=false;
		lig[i].spoExp=30.0f;
		lig[i].spoCut=5.0f;

		// Attenuation
		lig[i].conAtt=1.0f;
		lig[i].linAtt=0.0f;
		lig[i].quaAtt=0.0f;
	}

	// Switch light 0 and 7 on
	lig[0].setLig=lig[7].setLig=true;

	// Set parameters for spotlight 0
	lig[0].spoDir[0]= 0.10f;
	lig[0].spoDir[1]=-0.05f;
	lig[0].spoDir[2]=-0.60f;

	// Set all light-positions to positional light
	for(i=0; i<MAXLIG; i++) lig[i].ligPos[3]=1.0f;

	// Set "forground" light-positions
	lig[0].ligPos[0]=-1.0f; lig[0].ligPos[1]= 1.0f; lig[0].ligPos[2]= 1.0f;
	lig[1].ligPos[0]=-1.0f; lig[1].ligPos[1]=-1.0f; lig[1].ligPos[2]= 1.0f;
	lig[2].ligPos[0]= 1.0f; lig[2].ligPos[1]=-1.0f; lig[2].ligPos[2]= 1.0f;
	lig[3].ligPos[0]= 1.0f; lig[3].ligPos[1]= 1.0f; lig[3].ligPos[2]= 1.0f;

	// Set "background" light-positions
	lig[4].ligPos[0]=-1.0f; lig[4].ligPos[1]= 1.0f; lig[4].ligPos[2]=-5.0f;
	lig[5].ligPos[0]=-1.0f; lig[5].ligPos[1]=-1.0f; lig[5].ligPos[2]=-5.0f;
	lig[6].ligPos[0]= 1.0f; lig[6].ligPos[1]=-1.0f; lig[6].ligPos[2]=-5.0f;
	lig[7].ligPos[0]= 1.0f; lig[7].ligPos[1]= 1.0f; lig[7].ligPos[2]=-5.0f;

	// Set light-distances
	for(i=0; i<MAXLIG/2; i++) lig[i].ligDis=10;			// "Foreground"
	for(i=MAXLIG/2; i<MAXLIG; i++) lig[i].ligDis=50;	// "Background"
}


/* *************************************************** */
/* INITIALIZE MATERIALS WITH DEFAULT-VALUES            */
/* *************************************************** */
void iniMat() {
	int i, j, k;

	// Set all alpha-channels 1.0f
	for(k=0; k<2; k++)
		for(i=0; i<MAXFUN+1; i++) {
			mat[i].ambMat[k][3]=1.0f;
			mat[i].difMat[k][3]=1.0f;
			mat[i].speMat[k][3]=1.0f;
			mat[i].emiMat[k][3]=1.0f;

			mat[i].ambMatLoc[k]=false;
			mat[i].difMatLoc[k]=false;
			mat[i].speMatLoc[k]=false;
			mat[i].emiMatLoc[k]=false;
			mat[i].shiMat[k]=10.0f;
		}

	for(i=0; i<MAXFUN+1; i++)
		mat[i].twoSid=false;

	// Set common values for front-side
	for(i=0; i<MAXFUN+1; i++) {
		for(j=0; j<3; j++) mat[i].ambMat[0][j]=0.1f;
		for(j=0; j<3; j++) mat[i].difMat[0][j]=0.2f;
		for(j=0; j<3; j++) mat[i].speMat[0][j]=0.25f;
		for(j=0; j<3; j++) mat[i].emiMat[0][j]=0.0f;
	}

	// Set common values for back-side
	for(i=0; i<MAXFUN+1; i++) {
		for(j=0; j<3; j++) mat[i].ambMat[1][j]=0.1f;
		for(j=0; j<3; j++) mat[i].difMat[1][j]=0.9f;
		for(j=0; j<3; j++) mat[i].speMat[1][j]=0.25f;
		for(j=0; j<3; j++) mat[i].emiMat[1][j]=0.0f;
	}

	// Set default RGB-values
	mat[0].difMat[0][0]=0.9f;		// Function 0 red
	mat[1].difMat[0][1]=0.6f;		// Function 1 Green
	mat[2].difMat[0][2]=0.9f;		// Function 2 Blue
	mat[2].difMat[0][1]=0.35f;		// Function 2 Blue

	mat[3].difMat[0][0]=0.9f;		// Parametric
	mat[3].difMat[0][1]=0.4f;
}


/* *************************************************** */
/* INITIALIZE ENTIRE LIGHT WITH DEFAULT-VALUES         */
/* *************************************************** */
void iniEnt() {
	ent.ambEntLoc=true;		// Lock RGB-values
	ent.bacEntLoc=true;		// Lock RGB-values
	ent.modInf=1.0f;		// Default modell-view is local
	ent.modOne=0.0f;		// Default modell-view is 1-sided

	// Background light
	ent.bacLig[0]=0.2f;
	ent.bacLig[1]=0.2f;
	ent.bacLig[2]=0.2f;
	ent.bacLig[3]=1.0f;

	// Model ambient light
	ent.modLig[0]=0.0f;
	ent.modLig[1]=0.0f;
	ent.modLig[2]=0.0f;
	ent.modLig[3]=1.0f;
}


/* *************************************************** */
/* INITIALIZE ANIMATION WITH DEFAULT-VALUES            */
/* *************************************************** */
void iniAni() {
	ani.xFac=10;			// Factor for x-rotation; [0..100]
	ani.yFac=10;			// Factor for y-rotation; [0..100]
	ani.zFac=10;			// Factor for z-rotation; [0..100]
	ani.fps=50;				// Frames per second
	ani.locXYZ=false;		// X,Y,Z-sliders locked? bool
	ani.active=false;		// Not active; bool
}


/* *************************************************** */
/* INITIALIZE PIC-DIMENSIONS WITH DEFAULT-VALUES       */
/* *************************************************** */
void iniPic() {
	pic.xDim=0;				// Dummy-value, updated later
	pic.yDim=0;				// Dummy value, updated later
	pic.locXY=true;			// X/Y-values locked? bool
	pic.quality=85;			// Quality 0..100
	pic.png=true;			// PNG=default
}


/* *************************************************** */
/* INITIALIZE FUNCTIONS                                */
/* *************************************************** */
void iniFun() {
	q2cstrcpy(fun[0].str, "SIN(1.5*$PI*X*Y)+0.1");
	q2cstrcpy(fun[1].str, "XY*SIN(2.2*$PI*Y)-0.25");
	q2cstrcpy(fun[2].str, "X^2+XY-Y-10.5E-1");

	fun[0].drawFun=true;
	fun[1].drawFun=false;
	fun[2].drawFun=false;
}


/* *************************************************** */
/* INITIALIZE PARAMETRIC FUNCTIONS                     */
/* *************************************************** */
void iniPar() {
	q2cstrcpy(par.str[0], "-(0.7+0.3*COS(S))*SIN(T)");
	q2cstrcpy(par.str[1], "(0.7+0.3*COS(S))*COS(T)");
	q2cstrcpy(par.str[2], "0.3*SIN(S)");

	par.drawPar=false;
	par.sLow = par.tLow = -3.1416f;
	par.sHig = par.tHig =  3.1416f;
}


/* *************************************************** */
/* INITIALIZE STATUS                                   */
/* *************************************************** */
void iniSta() {
	sta.angle = 35.0;		// Initial angle for perspective

	sta.rotX = 0.0f;		// Initial values for rotations
	sta.rotY = 0.0f;
	sta.rotZ = 0.0f;

	sta.transX =  0.0f;		// Initial vValues For translations
	sta.transY =  0.0f;
	sta.transZ = -5.0f;

	sta.crossX = 0.5f;		// Initial coordinates for measuring-cross
	sta.crossY = 0.5f;

	sta.sa = 1.0f;			// Initial scale-factor over all
	sta.sx = 1.0f;			// Initial x,y,z scale-factors for the functions
	sta.sy = 1.0f;
	sta.sz = 1.0f;
	sta.spx = 1.0f;			// Initial x,y,z scale-factors for parametric view
	sta.spy = 1.0f;
	sta.spz = 1.0f;

	sta.tessWidth = 0.025f;	// Initial widht for the tesselations
	sta.grids=50;			// Initial number of gridlines

	sta.cross=false;		// Measuring-Cross
	sta.axes=true;			// Axes
	sta.wired=false;		// Polygon/full-mode
	sta.param=false;		// Parametric mode
	sta.draw=0;				// 0=Triangles, 1=Quads, 2=Points

	sta.ediLig=0;			// Edit light 0
	sta.ediMat=0;			// Edit function 0
	sta.parMod=false;		// Parameter mode off

	ani.xFac=ani.yFac=ani.zFac=10;
	ani.locXYZ=true;
	ani.fps=50;
	ani.active=false;
}


/* *************************************************** */
/* BACKUP-PROCEDURES                                   */
/* *************************************************** */
void saveToRam() {
	int i ,j, k;

	// Lights
	for(i=0; i<MAXLIG; i++) {
		for(j=0; j<4; j++) blig[i].ligPos[j] = lig[i].ligPos[j];
		for(j=0; j<4; j++) blig[i].ambLig[j] = lig[i].ambLig[j];
		for(j=0; j<4; j++) blig[i].difLig[j] = lig[i].difLig[j];
		for(j=0; j<4; j++) blig[i].speLig[j] = lig[i].speLig[j];
		for(j=0; j<4; j++) blig[i].spoDir[j] = lig[i].spoDir[j];

		blig[i].spoSta = lig[i].spoSta;
		blig[i].spoExp = lig[i].spoExp;
		blig[i].spoCut = lig[i].spoCut;
		blig[i].conAtt = lig[i].conAtt;
		blig[i].linAtt = lig[i].linAtt;
		blig[i].quaAtt = lig[i].quaAtt;
		blig[i].setLig = lig[i].setLig;
		blig[i].ambLigLoc = lig[i].ambLigLoc;
		blig[i].difLigLoc = lig[i].difLigLoc;
		blig[i].speLigLoc = lig[i].speLigLoc;
		blig[i].ligDis = lig[i].ligDis;
	}

	// Materials
	for(i=0; i<MAXFUN+1; i++) {
		for(j=0; j<4; j++)
			for(k=0; k<2; k++) {
				bmat[i].ambMat[k][j] = mat[i].ambMat[k][j];
				bmat[i].difMat[k][j] = mat[i].difMat[k][j];
				bmat[i].speMat[k][j] = mat[i].speMat[k][j];
				bmat[i].emiMat[k][j] = mat[i].emiMat[k][j];
			}
		for(k=0; k<2; k++) {
			bmat[i].ambMatLoc[k] = mat[i].ambMatLoc[k];
			bmat[i].difMatLoc[k] = mat[i].difMatLoc[k];
			bmat[i].speMatLoc[k] = mat[i].speMatLoc[k];
			bmat[i].emiMatLoc[k] = mat[i].emiMatLoc[k];
			bmat[i].shiMat[k] = mat[i].shiMat[k];
		}
		bmat[i].twoSid = mat[i].twoSid;
	}

	// Entire
	bent.ambEntLoc = ent.ambEntLoc;
	bent.bacEntLoc = ent.bacEntLoc;
	bent.modInf = ent.modInf;
	bent.modOne = ent.modOne;
	for(i=0; i<4; i++) {
		bent.bacLig[i] = ent.bacLig[i];
		bent.modLig[i] = ent.modLig[i];
	}

	// Animation
	bani.xFac = ani.xFac;
	bani.yFac = ani.yFac;
	bani.zFac = ani.zFac;
	bani.locXYZ = ani.locXYZ;
	bani.active = ani.active;
	bani.fps=ani.fps;

	// Status
	bsta.angle = sta.angle;
	bsta.rotX = sta.rotX;
	bsta.rotY = sta.rotY;
	bsta.rotZ = sta.rotZ;
	bsta.transX = sta.transX;
	bsta.transY = sta.transY;
	bsta.transZ = sta.transZ;
	bsta.crossX = sta.crossX;
	bsta.crossY = sta.crossY;
	bsta.sa = sta.sa;
	bsta.sx = sta.sx;
	bsta.sy = sta.sy;
	bsta.sz = sta.sz;
	bsta.spx = sta.spx;
	bsta.spy = sta.spy;
	bsta.spz = sta.spz;
	bsta.tessWidth = sta.tessWidth;
	bsta.grids = sta.grids;
	bsta.cross = sta.cross;
	bsta.axes = sta.axes;
	bsta.wired = sta.wired;
	bsta.param = sta.param;
	bsta.draw = sta.draw;
	bsta.ediLig = sta.ediLig;
	bsta.ediMat = sta.ediMat;
	bsta.parMod = sta.parMod;

	// Functions
	for(i=0; i<MAXFUN; i++) {
		strcpy(bfun[i].str, fun[i].str);
		strcpy(bpar.str[i], par.str[i]);
		bfun[i].drawFun = fun[i].drawFun;
	}
	bpar.drawPar = par.drawPar;
	bpar.sLow = par.sLow;
	bpar.sHig = par.sHig;
	bpar.tLow = par.tLow;
	bpar.tHig = par.tHig;
}


void restoreFromRam() {
	int i ,j, k;

	// Lights
	for(i=0; i<MAXLIG; i++) {
		for(j=0; j<4; j++) lig[i].ligPos[j] = blig[i].ligPos[j];
		for(j=0; j<4; j++) lig[i].ambLig[j] = blig[i].ambLig[j];
		for(j=0; j<4; j++) lig[i].difLig[j] = blig[i].difLig[j];
		for(j=0; j<4; j++) lig[i].speLig[j] = blig[i].speLig[j];
		for(j=0; j<4; j++) lig[i].spoDir[j] = blig[i].spoDir[j];

		lig[i].spoSta = blig[i].spoSta;
		lig[i].spoExp = blig[i].spoExp;
		lig[i].spoCut = blig[i].spoCut;
		lig[i].conAtt = blig[i].conAtt;
		lig[i].linAtt = blig[i].linAtt;
		lig[i].quaAtt = blig[i].quaAtt;
		lig[i].setLig = blig[i].setLig;
		lig[i].ambLigLoc = blig[i].ambLigLoc;
		lig[i].difLigLoc = blig[i].difLigLoc;
		lig[i].speLigLoc = blig[i].speLigLoc;
		lig[i].ligDis = blig[i].ligDis;
	}

	// Materials
	for(i=0; i<MAXFUN+1; i++) {
		for(j=0; j<4; j++)
			for(k=0; k<2; k++) {
				mat[i].ambMat[k][j] = bmat[i].ambMat[k][j];
				mat[i].difMat[k][j] = bmat[i].difMat[k][j];
				mat[i].speMat[k][j] = bmat[i].speMat[k][j];
				mat[i].emiMat[k][j] = bmat[i].emiMat[k][j];
			}
		for(k=0; k<2; k++) {
			mat[i].ambMatLoc[k] = bmat[i].ambMatLoc[k];
			mat[i].difMatLoc[k] = bmat[i].difMatLoc[k];
			mat[i].speMatLoc[k] = bmat[i].speMatLoc[k];
			mat[i].emiMatLoc[k] = bmat[i].emiMatLoc[k];
			mat[i].shiMat[k] = bmat[i].shiMat[k];
		}
		mat[i].twoSid = bmat[i].twoSid;
	}

	// Entire
	ent.ambEntLoc = bent.ambEntLoc;
	ent.bacEntLoc = bent.bacEntLoc;
	ent.modInf = bent.modInf;
	ent.modOne = bent.modOne;
	for(i=0; i<4; i++) {
		ent.bacLig[i] = bent.bacLig[i];
		ent.modLig[i] = bent.modLig[i];
	}

	// Animation
	ani.xFac = bani.xFac;
	ani.yFac = bani.yFac;
	ani.zFac = bani.zFac;
	ani.locXYZ = bani.locXYZ;
	ani.active = bani.active;
	ani.fps=bani.fps;

	// Status
	sta.angle = bsta.angle;
	sta.rotX = bsta.rotX;
	sta.rotY = bsta.rotY;
	sta.rotZ = bsta.rotZ;
	sta.transX = bsta.transX;
	sta.transY = bsta.transY;
	sta.transZ = bsta.transZ;
	sta.crossX = bsta.crossX;
	sta.crossY = bsta.crossY;
	sta.sa = bsta.sa;
	sta.sx = bsta.sx;
	sta.sy = bsta.sy;
	sta.sz = bsta.sz;
	sta.spx = bsta.spx;
	sta.spy = bsta.spy;
	sta.spz = bsta.spz;
	sta.tessWidth = bsta.tessWidth;
	sta.grids = bsta.grids;
	sta.cross = bsta.cross;
	sta.axes = bsta.axes;
	sta.wired = bsta.wired;
	sta.param = bsta.param;
	sta.draw = bsta.draw;
	sta.ediLig = bsta.ediLig;
	sta.ediMat = bsta.ediMat;
	sta.parMod = bsta.parMod;

	// Functions
	for(i=0; i<MAXFUN; i++) {
		strcpy(fun[i].str, bfun[i].str);
		strcpy(par.str[i], bpar.str[i]);
		fun[i].drawFun = bfun[i].drawFun;
	}
	par.drawPar = bpar.drawPar;
	par.sLow = bpar.sLow;
	par.sHig = bpar.sHig;
	par.tLow = bpar.tLow;
	par.tHig = bpar.tHig;
}
