/****************************************************************************
**
** Copyright (C) 1991-2006 Heinz van Saanen
**
** This file is part of the the function viewer Zhu3D.
**
** This file may be used under the terms of the GNU General Public
** License version 2.0 as published by the Free Software Foundation
** and appearing in the file LICENSE.GPL included in the packaging of
** this file.
**
** This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
** WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
**
****************************************************************************/

#ifndef PROPERTY_H
#define PROPERTY_H

#define MAXLIG 8			// Number of lights
#define MAXFUN 3			// Number of functions/materials
#define MAXLEN 128			// Maximum string lenght for functions

#include <QByteArray>
#include <QString>
#include <GL/gl.h>
#include "debug.h"


// Type-definitions for clarity
typedef GLfloat RGBA[4];	// Red,green,blue,alpha
typedef GLfloat XYZM[4];	// X,y,z,mode; mode==0=directional mode==1=position
typedef GLfloat XYZ[3];	// Coordinates


// Light property-type
typedef struct {
	int setLig;				// Individual light on?
	int ambLigLoc;			// Ambient RGB-sliders locked? bool
	int difLigLoc;			// Diffuse RGB-sliders locked? bool
	int speLigLoc;			// Specular RGB-sliders locked? bool

	int ligDis;				// Light distance; slider=[0,100]; default=10
	XYZM ligPos;			// Light position; internal=[0.0,10.0]; default=1.0

	RGBA ambLig;			// Ambient light
	RGBA difLig;			// Diffuse light
	RGBA speLig;			// Specular light (shininess)

	int     spoSta;			// Spotlight enabled?
	GLfloat spoExp;			// Spotlight intensity; slider=[0,45]; internal={0,135]
	GLfloat spoCut;			// Spotlight cutoff-angle; slider=internal=[0,45]; 180=off
	XYZ     spoDir;			// Spotlight direction-vector

	GLfloat conAtt;			// Constant attenuation; slider=[0,255]; internal=[0.0,2.55]
	GLfloat linAtt;			// Linear attenuation
	GLfloat quaAtt;			// Quadratic attenuation
} LIG;


// Material property-type
// mat[0]..front-side; mat[1]..back-side
typedef struct {
	int ambMatLoc[2];		// Ambient RGB-sliders locked? bool
	int difMatLoc[2];		// Diffuse RGB-sliders locked? bool
	int speMatLoc[2];		// Specular RGB-sliders locked? bool
	int emiMatLoc[2];		// Emission RBG-sliders locked? bool
	RGBA ambMat[2];			// Ambient material
	RGBA difMat[2];			// Diffuse material
	RGBA speMat[2];			// Specular material for shininess
	RGBA emiMat[2];			// Emission material
	GLfloat shiMat[2];		// Shininess; slider=internal=[0,128]
	int twoSid;				// Back-side button set? bool
} MAT;


// Entire OpenGL-property-type
typedef struct {
	int ambEntLoc;			// Ambient RGB-sliders locked? bool
	int bacEntLoc;			// Background RGB-sliders locked? bool
	RGBA modLig;			// Entire ambient RGBA-light
	RGBA bacLig;			// Entire background colour
	GLfloat modInf;			// Modes: infinite=0.0 local=1.0; defaul=local
	GLfloat modOne;			// Modes: 1-sided=0.0 2-sided=1.0; default=1-sided
} ENT;


// Animation-property-type
typedef struct {
	int xFac;			// Factor for x-rotation; [0..100]
	int yFac;			// Factor for y-rotation; [0..100]
	int zFac;			// Factor for z-rotation; [0..100]
	int fps;			// Frames per second
	int locXYZ;			// X,Y,Z-sliders locked? bool
	int active;			// Currently active? bool
} ANI;


// The function-type
typedef struct {
	char str[MAXLEN];		// Plain old Ascii-string
	int drawFun;			// Function to draw; bool
} FUN;


// The parameter-type
typedef struct {
	char str[3][MAXLEN];	// Strings for u(s,t), v(s,t), w(s,t)
	int drawPar;			// Parametric view to draw; bool
	float sLow, sHig;		// Interval: sLow < s < sHig
	float tLow, tHig;		// Interval: tLow < t < tHig
} PAR;


// Status value-type
typedef struct {
	GLdouble angle;			// Viewing angle

	GLfloat rotX;			// For rotations
	GLfloat rotY;
	GLfloat rotZ;
	GLfloat transX;			// For translations
	GLfloat transY;
	GLfloat transZ;
	GLfloat crossX;			// For cross-coordinates
	GLfloat crossY;
	GLfloat crossZ;

	float sa;				// Initial scale-factor over all
	float sx;				// Initial x,y,z scale-factors for the functions
	float sy;
	float sz;
	float spx;				// Initial x,y,z scale-factors parametric view
	float spy;
	float spz;

	float tessWidth;		// Initial widht for the tesselations
	int grids;				// Initial number of gridlines

	int cross;				// Show Measuring-Cross; bool
	int axes;				// Show Axes; bool
	int wired;				// Polygon/full-mode; bool
	int param;				// Parametric mode; bool
	int draw;				// 0=Triangles, 1=Quads, 2=Points

	int ediLig;				// Holds actual edited light 0..7
	int ediMat;				// Holds actual edited material 0..2
	int parMod;				// Parameters mode; bool
} STA;


// The picture-type
typedef struct {
	int xDim;		// x-dimension for snapshot
	int yDim;		// y-dimension for snapshot
	int locXY;		// x-y locked?; bool
	int quality;	// Quality 0..100
	int png;		// PNG-format? bool
} PIC;


// Constant-type
typedef struct {
	float zoomFac;			// Determines zoom-factor
	float transFac;			// Determines translation
	float fineFac;			// Determines fine scaling-factor
	float normalFac;		// Determines normal scaling-factor
	float coarseFac;		// Determines coarse scaling-factor
	float tessFac;			// Determines tesselation change
} CON;


// Global access
extern LIG lig[MAXLIG];
extern MAT mat[MAXFUN+1];
extern ENT ent;
extern ANI ani;
extern FUN fun[MAXFUN];
extern PAR par;
extern STA sta;
extern PIC pic;
extern const CON con;


// Functions & procedures
void iniLig();
void iniMat();
void iniEnt();
void iniAni();
void iniFun();
void iniPar();
void iniSta();
void iniPic();


// Miscellanous helpers
void q2cstrcpy(char *str, const QString qstr);
void saveToRam();
void restoreFromRam();

#endif
